from autoconf import conf

dir(conf)
from autoarray import exc
from autoarray import masked
from autoarray import plot
from autoarray import simulator
from autoarray import util
from autoarray.dataset import data_converter
from autoarray.dataset.imaging import Imaging as imaging
from autoarray.dataset.interferometer import Interferometer as interferometer
from autoarray.fit.fit import fit
from autoarray.mask.mask import Mask as mask
from autoarray.operators.convolver import Convolver as convolver
from autoarray.operators.inversion import pixelizations as pix, regularization as reg
from autoarray.operators.inversion.inversions import inversion
from autoarray.operators.inversion.mappers import mapper
from autoarray.operators.transformer import Transformer as transformer
from autoarray.structures.arrays import Array as array
from autoarray.structures.grids import (
    Grid as grid,
    GridIrregular as grid_irregular,
    GridRectangular as grid_rectangular,
    GridVoronoi as grid_voronoi,
    Coordinates as coordinates,
)
from autoarray.structures.kernel import Kernel as kernel
from autoarray.structures.visibilities import Visibilities as visibilities

__version__ = '0.7.9'



import logging

import numpy as np

from autoarray import exc
from autoarray.structures.arrays import AbstractArray
from autoarray.structures.grids import AbstractGrid
from autoarray.util import array_util, grid_util

logging.basicConfig()
logger = logging.getLogger(__name__)


class MaskedArray(AbstractArray):
    @classmethod
    def manual_1d(cls, array, mask, store_in_1d=True):

        if type(array) is list:
            array = np.asarray(array)

        if array.shape[0] != mask.sub_pixels_in_mask:
            raise exc.ArrayException(
                "The input 1D array does not have the same number of entries as sub-pixels in"
                "the mask."
            )

        if store_in_1d:
            return mask.mapping.array_stored_1d_from_sub_array_1d(sub_array_1d=array)
        else:
            return mask.mapping.array_stored_2d_from_sub_array_1d(sub_array_1d=array)

    @classmethod
    def manual_2d(cls, array, mask, store_in_1d=True):

        if type(array) is list:
            array = np.asarray(array)

        if array.shape != mask.sub_shape_2d:
            raise exc.ArrayException(
                "The input array is 2D but not the same dimensions as the sub-mask "
                "(e.g. the mask 2D shape multipled by its sub size."
            )

        if store_in_1d:
            return mask.mapping.array_stored_1d_from_sub_array_2d(sub_array_2d=array)
        else:
            masked_sub_array_1d = mask.mapping.array_stored_1d_from_sub_array_2d(
                sub_array_2d=array
            )
            return mask.mapping.array_stored_2d_from_sub_array_1d(
                sub_array_1d=masked_sub_array_1d
            )

    @classmethod
    def full(cls, fill_value, mask, store_in_1d=True):
        return cls.manual_2d(
            array=np.full(fill_value=fill_value, shape=mask.sub_shape_2d),
            mask=mask,
            store_in_1d=store_in_1d,
        )

    @classmethod
    def ones(cls, mask, store_in_1d=True):
        return cls.full(fill_value=1.0, mask=mask, store_in_1d=store_in_1d)

    @classmethod
    def zeros(cls, mask, store_in_1d=True):
        return cls.full(fill_value=0.0, mask=mask, store_in_1d=store_in_1d)

    @classmethod
    def from_fits(cls, file_path, hdu, mask, store_in_1d=True):
        array_2d = array_util.numpy_array_2d_from_fits(file_path=file_path, hdu=hdu)
        return cls.manual_2d(array=array_2d, mask=mask, store_in_1d=store_in_1d)


class MaskedGrid(AbstractGrid):
    @classmethod
    def manual_1d(cls, grid, mask, store_in_1d=True):

        if type(grid) is list:
            grid = np.asarray(grid)

        if grid.shape[0] != mask.sub_pixels_in_mask:
            raise exc.GridException(
                "The input 1D grid does not have the same number of entries as sub-pixels in"
                "the mask."
            )

        if store_in_1d:
            return mask.mapping.grid_stored_1d_from_sub_grid_1d(sub_grid_1d=grid)
        else:
            return mask.mapping.grid_stored_2d_from_sub_grid_1d(sub_grid_1d=grid)

    @classmethod
    def manual_2d(cls, grid, mask, store_in_1d=True):

        if type(grid) is list:
            grid = np.asarray(grid)

        if (grid.shape[0], grid.shape[1]) != mask.sub_shape_2d:
            raise exc.GridException(
                "The input grid is 2D but not the same dimensions as the sub-mask "
                "(e.g. the mask 2D shape multipled by its sub size."
            )

        if store_in_1d:
            return mask.mapping.grid_stored_1d_from_sub_grid_2d(sub_grid_2d=grid)
        else:
            sub_grid_1d = mask.mapping.grid_stored_1d_from_sub_grid_2d(sub_grid_2d=grid)
            return mask.mapping.grid_stored_2d_from_sub_grid_1d(sub_grid_1d=sub_grid_1d)

    @classmethod
    def from_mask(cls, mask, store_in_1d=True):
        """Setup a sub-grid of the unmasked pixels, using a mask and a specified sub-grid size. The center of \
        every unmasked pixel's sub-pixels give the grid's (y,x) arc-second coordinates.

        Parameters
        -----------
        mask : Mask
            The mask whose masked pixels are used to setup the sub-pixel grid.
        sub_size : int
            The size (sub_size x sub_size) of each unmasked pixels sub-grid.
        """

        sub_grid_1d = grid_util.grid_1d_via_mask_2d(
            mask_2d=mask,
            pixel_scales=mask.pixel_scales,
            sub_size=mask.sub_size,
            origin=mask.origin,
        )

        if store_in_1d:
            return mask.mapping.grid_stored_1d_from_sub_grid_1d(sub_grid_1d=sub_grid_1d)
        else:
            return mask.mapping.grid_stored_2d_from_sub_grid_1d(sub_grid_1d=sub_grid_1d)



import logging
import numpy as np

import autoarray as aa

from autoarray import exc
from autoarray.dataset import abstract_dataset
from autoarray.structures import visibilities as vis


logger = logging.getLogger(__name__)


class AbstractInterferometerSet(abstract_dataset.AbstractDataset):
    @property
    def visibilities(self):
        return self.data

    @property
    def amplitudes(self):
        return self.visibilities.amplitudes

    @property
    def phases(self):
        return self.visibilities.phases

    @property
    def uv_distances(self):
        return np.sqrt(
            np.square(self.uv_wavelengths[:, 0]) + np.square(self.uv_wavelengths[:, 1])
        )

    def modified_visibilities_from_visibilities(self, visibilities):

        return Interferometer(
            visibilities=visibilities,
            noise_map=self.noise_map,
            uv_wavelengths=self.uv_wavelengths,
            primary_beam=self.primary_beam,
            exposure_time_map=self.exposure_time_map,
        )

    def resized_primary_beam_from_new_shape_2d(self, new_shape_2d):

        primary_beam = self.primary_beam.resized_from_new_shape(new_shape=new_shape_2d)
        return Interferometer(
            visibilities=self.data,
            noise_map=self.noise_map,
            exposure_time_map=self.exposure_time_map,
            uv_wavelengths=self.uv_wavelengths,
            primary_beam=primary_beam,
        )

    def output_to_fits(
        self,
        visibilities_path=None,
        noise_map_path=None,
        primary_beam_path=None,
        exposure_time_map_path=None,
        uv_wavelengths_path=None,
        overwrite=False,
    ):

        if primary_beam_path is not None:
            self.primary_beam.output_to_fits(
                file_path=primary_beam_path, overwrite=overwrite
            )

        if self.exposure_time_map is not None and exposure_time_map_path is not None:
            aa.util.array.numpy_array_1d_to_fits(
                array_1d=self.exposure_time_map,
                file_path=exposure_time_map_path,
                overwrite=overwrite,
            )

        if visibilities_path is not None:
            self.visibilities.output_to_fits(
                file_path=visibilities_path, overwrite=overwrite
            )

        if self.noise_map is not None and noise_map_path is not None:
            self.noise_map.output_to_fits(file_path=noise_map_path, overwrite=overwrite)

        if self.uv_wavelengths is not None and uv_wavelengths_path is not None:
            aa.util.array.numpy_array_2d_to_fits(
                array_2d=self.uv_wavelengths,
                file_path=uv_wavelengths_path,
                overwrite=overwrite,
            )


class Interferometer(AbstractInterferometerSet):
    def __init__(
        self,
        visibilities,
        noise_map,
        uv_wavelengths,
        primary_beam=None,
        exposure_time_map=None,
    ):

        super(Interferometer, self).__init__(
            data=visibilities, noise_map=noise_map, exposure_time_map=exposure_time_map
        )

        self.uv_wavelengths = uv_wavelengths
        self.primary_beam = primary_beam

    @classmethod
    def manual(
        cls,
        visibilities,
        noise_map,
        uv_wavelengths,
        primary_beam=None,
        exposure_time_map=None,
    ):
        return Interferometer(
            visibilities=visibilities,
            noise_map=noise_map,
            uv_wavelengths=uv_wavelengths,
            primary_beam=primary_beam,
            exposure_time_map=exposure_time_map,
        )

    @classmethod
    def from_fits(
        cls,
        visibilities_path,
        noise_map_path,
        uv_wavelengths_path,
        visibilities_hdu=0,
        noise_map_hdu=0,
        uv_wavelengths_hdu=0,
        resized_primary_beam_shape_2d=None,
        renormalize_primary_beam=True,
        exposure_time_map_path=None,
        exposure_time_map_hdu=0,
        exposure_time_map_from_single_value=None,
        primary_beam_path=None,
        primary_beam_hdu=0,
    ):
        """Factory for loading the interferometer data_type from .fits files, as well as computing properties like the noise-map,
        exposure-time map, etc. from the interferometer-data_type.

        This factory also includes a number of routines for converting the interferometer-data_type from unit_label not supported by PyAutoLens \
        (e.g. adus, electrons) to electrons per second.

        Parameters
        ----------
        lens_name
        image_path : str
            The path to the image .fits file containing the image (e.g. '/path/to/image.fits')
        image_hdu : int
            The hdu the image is contained in the .fits file specified by *image_path*.
        image_hdu : int
            The hdu the image is contained in the .fits file that *image_path* points too.
        resized_interferometer_shape : (int, int) | None
            If input, the interferometer structures that are image sized, e.g. the image, noise-maps) are resized to these dimensions.
        resized_interferometer_origin_pixels : (int, int) | None
            If the interferometer structures are resized, this defines a new origin (in pixels) around which recentering occurs.
        resized_interferometer_origin_arcsec : (float, float) | None
            If the interferometer structures are resized, this defines a new origin (in arc-seconds) around which recentering occurs.
        primary_beam_path : str
            The path to the primary_beam .fits file containing the primary_beam (e.g. '/path/to/primary_beam.fits')
        primary_beam_hdu : int
            The hdu the primary_beam is contained in the .fits file specified by *primary_beam_path*.
        resized_primary_beam_shape_2d : (int, int) | None
            If input, the primary_beam is resized to these dimensions.
        renormalize_psf : bool
            If True, the PrimaryBeam is renoralized such that all elements sum to 1.0.
        noise_map_path : str
            The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')
        noise_map_hdu : int
            The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
        noise_map_from_image_and_background_noise_map : bool
            If True, the noise-map is computed from the observed image and background noise-map \
            (see NoiseMap.from_image_and_background_noise_map).
        convert_noise_map_from_weight_map : bool
            If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
            *NoiseMap.from_weight_map).
        convert_noise_map_from_inverse_noise_map : bool
            If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
            *NoiseMap.from_inverse_noise_map).
        background_noise_map_path : str
            The path to the background_noise_map .fits file containing the background noise-map \
            (e.g. '/path/to/background_noise_map.fits')
        background_noise_map_hdu : int
            The hdu the background_noise_map is contained in the .fits file specified by *background_noise_map_path*.
        convert_background_noise_map_from_weight_map : bool
            If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
            *NoiseMap.from_weight_map).
        convert_background_noise_map_from_inverse_noise_map : bool
            If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
            noise-map (see *NoiseMap.from_inverse_noise_map).
        poisson_noise_map_path : str
            The path to the poisson_noise_map .fits file containing the Poisson noise-map \
             (e.g. '/path/to/poisson_noise_map.fits')
        poisson_noise_map_hdu : int
            The hdu the poisson_noise_map is contained in the .fits file specified by *poisson_noise_map_path*.
        poisson_noise_map_from_image : bool
            If True, the Poisson noise-map is estimated using the image.
        convert_poisson_noise_map_from_weight_map : bool
            If True, the Poisson noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
            *NoiseMap.from_weight_map).
        convert_poisson_noise_map_from_inverse_noise_map : bool
            If True, the Poisson noise-map loaded from the .fits file is converted from an inverse noise-map to a \
            noise-map (see *NoiseMap.from_inverse_noise_map).
        exposure_time_map_path : str
            The path to the exposure_time_map .fits file containing the exposure time map \
            (e.g. '/path/to/exposure_time_map.fits')
        exposure_time_map_hdu : int
            The hdu the exposure_time_map is contained in the .fits file specified by *exposure_time_map_path*.
        exposure_time_map_from_single_value : float
            The exposure time of the interferometer imaging, which is used to compute the exposure-time map as a single value \
            (see *ExposureTimeMap.from_single_value*).
        exposure_time_map_from_inverse_noise_map : bool
            If True, the exposure-time map is computed from the background noise_map map \
            (see *ExposureTimeMap.from_background_noise_map*)
        background_sky_map_path : str
            The path to the background_sky_map .fits file containing the background sky map \
            (e.g. '/path/to/background_sky_map.fits').
        background_sky_map_hdu : int
            The hdu the background_sky_map is contained in the .fits file specified by *background_sky_map_path*.
        convert_from_electrons : bool
            If True, the input unblurred_image_1d are in unit_label of electrons and all converted to electrons / second using the exposure \
            time map.
        gain : float
            The image gain, used for convert from ADUs.
        convert_from_adus : bool
            If True, the input unblurred_image_1d are in unit_label of adus and all converted to electrons / second using the exposure \
            time map and gain.
        """

        visibilities = aa.visibilities.from_fits(
            file_path=visibilities_path, hdu=visibilities_hdu
        )

        exposure_time_map = load_exposure_time_map(
            exposure_time_map_path=exposure_time_map_path,
            exposure_time_map_hdu=exposure_time_map_hdu,
            shape_1d=visibilities.shape_1d,
            exposure_time=exposure_time_map_from_single_value,
        )

        noise_map = aa.visibilities.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu
        )

        uv_wavelengths = aa.util.array.numpy_array_2d_from_fits(
            file_path=uv_wavelengths_path, hdu=uv_wavelengths_hdu
        )

        if primary_beam_path is not None:
            primary_beam = aa.kernel.from_fits(
                file_path=primary_beam_path,
                hdu=primary_beam_hdu,
                renormalize=renormalize_primary_beam,
            )
        else:
            primary_beam = None

        interferometer = Interferometer(
            visibilities=visibilities,
            primary_beam=primary_beam,
            noise_map=noise_map,
            uv_wavelengths=uv_wavelengths,
            exposure_time_map=exposure_time_map,
        )

        if resized_primary_beam_shape_2d is not None:
            interferometer = interferometer.resized_primary_beam_from_new_shape_2d(
                new_shape_2d=resized_primary_beam_shape_2d
            )

        return interferometer

    @classmethod
    def simulate(
        cls,
        real_space_image,
        real_space_pixel_scales,
        exposure_time,
        transformer,
        primary_beam=None,
        exposure_time_map=None,
        background_level=0.0,
        background_sky_map=None,
        noise_sigma=None,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        real_space_image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and UVPlane read-out).
        real_space_pixel_scales: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        if type(real_space_pixel_scales) is float:
            real_space_pixel_scales = (real_space_pixel_scales, real_space_pixel_scales)

        if exposure_time_map is None:

            exposure_time_map = aa.array.full(
                fill_value=exposure_time,
                shape_2d=real_space_image.shape_2d,
                pixel_scales=real_space_pixel_scales,
            )

        if background_sky_map is None:

            background_sky_map = aa.array.full(
                fill_value=background_level,
                shape_2d=real_space_image.shape_2d,
                pixel_scales=real_space_pixel_scales,
            )

        real_space_image += background_sky_map

        visibilities = transformer.visibilities_from_image(image=real_space_image)

        if noise_sigma is not None:
            noise_map_realization = gaussian_noise_map_from_shape_and_sigma(
                shape=visibilities.shape, sigma=noise_sigma, noise_seed=noise_seed
            )
            visibilities = visibilities + noise_map_realization
            noise_map = np.full(fill_value=noise_sigma, shape=visibilities.shape)
        else:
            noise_map = np.full(
                fill_value=noise_if_add_noise_false, shape=visibilities.shape
            )
            noise_map_realization = None

        noise_map = vis.Visibilities.manual_1d(visibilities=noise_map)

        if np.isnan(noise_map).any():
            raise exc.DataException(
                "The noise-map has NaN values in it. This suggests your exposure time and / or"
                "background sky levels are too low, creating signal counts at or close to 0.0."
            )

        real_space_image -= background_sky_map

        return SimulatedInterferometer(
            real_space_shape_2d=real_space_image.shape,
            visibilities=visibilities,
            real_space_pixel_scales=real_space_pixel_scales,
            noise_map=noise_map,
            uv_wavelengths=transformer.uv_wavelengths,
            primary_beam=primary_beam,
            noise_map_realization=noise_map_realization,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            noise_realization=noise_map_realization,
        )


class SimulatedInterferometer(Interferometer):
    def __init__(
        self,
        real_space_shape_2d,
        real_space_pixel_scales,
        visibilities,
        noise_map,
        uv_wavelengths,
        primary_beam,
        noise_map_realization,
        exposure_time_map=None,
        **kwargs
    ):

        super(SimulatedInterferometer, self).__init__(
            visibilities=visibilities,
            noise_map=noise_map,
            uv_wavelengths=uv_wavelengths,
            primary_beam=primary_beam,
            exposure_time_map=exposure_time_map,
        )

        self.real_space_shape_2d = real_space_shape_2d
        self.real_space_pixel_scales = real_space_pixel_scales
        self.noise_map_realization = noise_map_realization

    def __array_finalize__(self, obj):
        if isinstance(obj, SimulatedInterferometer):
            try:
                self.data = obj.data
                self.real_space_shape_2d = obj.real_space_shape_2d
                self.real_space_pixel_scales = obj.real_space_pixel_scales
                self.magnitudes = obj.magnitudes
                self.noise_map = obj.noise_map
                self.noise_map_realization = obj.noise_map_realization
                self.primary_beam = obj.primary_beam
                self.exposure_time_map = obj.exposure_time_map
                self.origin = obj.origin
            except AttributeError:
                logger.debug(
                    "Original object in UVPlane.__array_finalize__ missing one or more attributes"
                )


def gaussian_noise_map_from_shape_and_sigma(shape, sigma, noise_seed=-1):
    """Generate a two-dimensional read noises-map, generating values from a Gaussian distribution with mean 0.0.

    Params
    ----------
    shape : (int, int)
        The (x,y) image_shape of the generated Gaussian noises map.
    read_noise : float
        Standard deviation of the 1D Gaussian that each noises value is drawn from
    seed : int
        The seed of the random number generator, used for the random noises maps.
    """
    if noise_seed == -1:
        # Use one seed, so all regions have identical column non-uniformity.
        noise_seed = np.random.randint(0, int(1e9))
    np.random.seed(noise_seed)
    read_noise_map = np.random.normal(loc=0.0, scale=sigma, size=shape)
    return read_noise_map


def load_exposure_time_map(
    exposure_time_map_path, exposure_time_map_hdu, shape_1d=None, exposure_time=None
):
    """Factory for loading the exposure time map from a .fits file.

    This factory also includes a number of routines for computing the exposure-time map from other unblurred_image_1d \
    (e.g. the background noise-map).

    Parameters
    ----------
    exposure_time_map_path : str
        The path to the exposure_time_map .fits file containing the exposure time map \
        (e.g. '/path/to/exposure_time_map.fits')
    exposure_time_map_hdu : int
        The hdu the exposure_time_map is contained in the .fits file specified by *exposure_time_map_path*.
    pixel_scales : float
        The size of each pixel in arc seconds.
    shape_1d : (int, int)
        The shape of the image, required if a single value is used to calculate the exposure time map.
    exposure_time : float
        The exposure-time used to compute the expsure-time map if only a single value is used.
    exposure_time_map_from_inverse_noise_map : bool
        If True, the exposure-time map is computed from the background noise_map map \
        (see *ExposureTimeMap.from_background_noise_map*)
    inverse_noise_map : ndarray
        The background noise-map, which the Poisson noise-map can be calculated using.
    """

    if exposure_time is not None and exposure_time_map_path is not None:
        raise exc.DataException(
            "You have supplied both a exposure_time_map_path to an exposure time map and an exposure time. Only"
            "one quantity should be supplied."
        )

    if exposure_time is not None and exposure_time_map_path is None:
        return np.full(fill_value=exposure_time, shape=shape_1d)
    elif exposure_time is None and exposure_time_map_path is not None:
        return aa.util.array.numpy_array_1d_from_fits(
            file_path=exposure_time_map_path, hdu=exposure_time_map_hdu
        )



import os

from autoarray.util import array_util
from autoarray.structures import grids, kernel
from autoarray.dataset import imaging, interferometer
from autoarray.operators import transformer


class ImagingSimulator:
    def __init__(
        self,
        shape_2d,
        pixel_scales,
        sub_size,
        psf,
        exposure_time,
        background_level,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        origin=(0.0, 0.0),
    ):
        """A class representing a Imaging observation, using the shape of the image, the pixel scale,
        psf, exposure time, etc.

        Parameters
        ----------
        shape_2d : (int, int)
            The shape of the observation. Note that we do not simulator a full Imaging frame (e.g. 2000 x 2000 pixels for \
            Hubble imaging), but instead just a cut-out around the strong lens.
        pixel_scales : float
            The size of each pixel in arc seconds.
        psf : PSF
            An arrays describing the PSF kernel of the image.
        exposure_time : float
            The exposure time of an observation using this data_type.
        background_level : float
            The level of the background sky of an observationg using this data_type.
        """

        if type(pixel_scales) is float:
            pixel_scales = (pixel_scales, pixel_scales)

        self.shape_2d = shape_2d
        self.pixel_scales = pixel_scales
        self.sub_size = sub_size
        self.origin = origin
        self.psf = psf
        self.exposure_time = exposure_time
        self.background_level = background_level
        self.add_noise = add_noise
        self.noise_if_add_noise_false = noise_if_add_noise_false
        self.noise_seed = noise_seed

    @classmethod
    def lsst(
        cls,
        shape=(101, 101),
        pixel_scales=0.2,
        sub_size=8,
        psf_shape_2d=(31, 31),
        psf_sigma=0.5,
        exposure_time=100.0,
        background_level=1.0,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """Default settings for an observation with the Large Synotpic Survey Telescope.

        This can be customized by over-riding the default input values."""
        psf = kernel.Kernel.from_gaussian(
            shape_2d=psf_shape_2d, sigma=psf_sigma, pixel_scales=pixel_scales
        )
        return cls(
            shape_2d=shape,
            pixel_scales=pixel_scales,
            sub_size=sub_size,
            psf=psf,
            exposure_time=exposure_time,
            background_level=background_level,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    @classmethod
    def euclid(
        cls,
        shape=(151, 151),
        pixel_scales=0.1,
        sub_size=8,
        psf_shape_2d=(31, 31),
        psf_sigma=0.1,
        exposure_time=565.0,
        background_level=1.0,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """Default settings for an observation with the Euclid space satellite.

        This can be customized by over-riding the default input values."""
        psf = kernel.Kernel.from_gaussian(
            shape_2d=psf_shape_2d, sigma=psf_sigma, pixel_scales=pixel_scales
        )
        return cls(
            shape_2d=shape,
            pixel_scales=pixel_scales,
            sub_size=sub_size,
            psf=psf,
            exposure_time=exposure_time,
            background_level=background_level,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    @classmethod
    def hst(
        cls,
        shape=(251, 251),
        pixel_scales=0.05,
        sub_size=8,
        psf_shape_2d=(31, 31),
        psf_sigma=0.05,
        exposure_time=2000.0,
        background_level=1.0,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """Default settings for an observation with the Hubble Space Telescope.

        This can be customized by over-riding the default input values."""
        psf = kernel.Kernel.from_gaussian(
            shape_2d=psf_shape_2d, sigma=psf_sigma, pixel_scales=pixel_scales
        )
        return cls(
            shape_2d=shape,
            pixel_scales=pixel_scales,
            sub_size=sub_size,
            psf=psf,
            exposure_time=exposure_time,
            background_level=background_level,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    @classmethod
    def hst_up_sampled(
        cls,
        shape=(401, 401),
        pixel_scales=0.03,
        sub_size=8,
        psf_shape_2d=(31, 31),
        psf_sigma=0.05,
        exposure_time=2000.0,
        background_level=1.0,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """Default settings for an observation with the Hubble Space Telescope which has been upscaled to a higher \
        pixel-scale to better sample the PSF.

        This can be customized by over-riding the default input values."""
        psf = kernel.Kernel.from_gaussian(
            shape_2d=psf_shape_2d, sigma=psf_sigma, pixel_scales=pixel_scales
        )
        return cls(
            shape_2d=shape,
            pixel_scales=pixel_scales,
            sub_size=sub_size,
            psf=psf,
            exposure_time=exposure_time,
            background_level=background_level,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    @classmethod
    def keck_adaptive_optics(
        cls,
        shape=(751, 751),
        pixel_scales=0.01,
        sub_size=8,
        psf_shape_2d=(31, 31),
        psf_sigma=0.025,
        exposure_time=1000.0,
        background_level=1.0,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """Default settings for an observation using Keck Adaptive Optics imaging.

        This can be customized by over-riding the default input values."""
        psf = kernel.Kernel.from_gaussian(
            shape_2d=psf_shape_2d, sigma=psf_sigma, pixel_scales=pixel_scales
        )
        return cls(
            shape_2d=shape,
            pixel_scales=pixel_scales,
            sub_size=sub_size,
            psf=psf,
            exposure_time=exposure_time,
            background_level=background_level,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    @property
    def grid(self):
        return grids.Grid.uniform(
            shape_2d=self.shape_2d,
            pixel_scales=self.pixel_scales,
            sub_size=self.sub_size,
            origin=self.origin,
        )

    def from_image(self, image, name=None):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Imaging read-out).
        pixel_scales: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An arrays representing the effective exposure time of each pixel.
        psf: PSF
            An arrays describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """
        return imaging.SimulatedImaging.simulate(
            image=image,
            exposure_time=self.exposure_time,
            psf=self.psf,
            background_level=self.background_level,
            add_noise=self.add_noise,
            noise_if_add_noise_false=self.noise_if_add_noise_false,
            noise_seed=self.noise_seed,
            name=name,
        )


class InterferometerSimulator:
    def __init__(
        self,
        real_space_shape_2d,
        real_space_pixel_scales,
        uv_wavelengths,
        sub_size,
        exposure_time,
        background_level,
        primary_beam=None,
        noise_sigma=0.1,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        origin=(0.0, 0.0),
    ):
        """A class representing a Imaging observation, using the shape of the image, the pixel scale,
        psf, exposure time, etc.

        Parameters
        ----------
        real_space_shape_2d : (int, int)
            The shape of the observation. Note that we do not simulator a full Imaging frame (e.g. 2000 x 2000 pixels for \
            Hubble imaging), but instead just a cut-out around the strong lens.
        real_space_pixel_scales : float
            The size of each pixel in arc seconds.
        psf : PSF
            An arrays describing the PSF kernel of the image.
        exposure_time : float
            The exposure time of an observation using this data_type.
        background_level : float
            The level of the background sky of an observationg using this data_type.
        """

        if type(real_space_pixel_scales) is float:
            real_space_pixel_scales = (real_space_pixel_scales, real_space_pixel_scales)

        self.real_space_shape_2d = real_space_shape_2d
        self.real_space_pixel_scales = real_space_pixel_scales
        self.uv_wavelengths = uv_wavelengths
        self.sub_size = sub_size
        self.origin = origin
        self.transformer = transformer.Transformer(
            uv_wavelengths=self.uv_wavelengths,
            grid_radians=self.grid.in_1d_binned.in_radians,
        )
        self.exposure_time = exposure_time
        self.background_level = background_level
        self.primary_beam = primary_beam
        self.noise_sigma = noise_sigma
        self.noise_if_add_noise_false = noise_if_add_noise_false
        self.noise_seed = noise_seed

    @property
    def grid(self):
        return grids.Grid.uniform(
            shape_2d=self.real_space_shape_2d,
            pixel_scales=self.real_space_pixel_scales,
            sub_size=self.sub_size,
            origin=self.origin,
        )

    @classmethod
    def sma(
        cls,
        real_space_shape_2d=(151, 151),
        real_space_pixel_scales=(0.05, 0.05),
        sub_size=8,
        primary_beam_shape_2d=None,
        primary_beam_sigma=None,
        exposure_time=100.0,
        background_level=1.0,
        noise_sigma=0.1,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """Default settings for an observation with the Large Synotpic Survey Telescope.

        This can be customized by over-riding the default input values."""

        uv_wavelengths_path = "{}/dataset/sma_uv_wavelengths.fits".format(
            os.path.dirname(os.path.realpath(__file__))
        )

        uv_wavelengths = array_util.numpy_array_1d_from_fits(
            file_path=uv_wavelengths_path, hdu=0
        )

        if primary_beam_shape_2d is not None and primary_beam_sigma is not None:
            primary_beam = kernel.Kernel.from_gaussian(
                shape_2d=primary_beam_shape_2d,
                sigma=primary_beam_sigma,
                pixel_scales=real_space_pixel_scales,
            )
        else:
            primary_beam = None

        return cls(
            real_space_shape_2d=real_space_shape_2d,
            real_space_pixel_scales=real_space_pixel_scales,
            uv_wavelengths=uv_wavelengths,
            sub_size=sub_size,
            primary_beam=primary_beam,
            exposure_time=exposure_time,
            background_level=background_level,
            noise_sigma=noise_sigma,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    def from_real_space_image(self, real_space_image):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        real_space_image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Imaging read-out).
        pixel_scales: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An arrays representing the effective exposure time of each pixel.
        psf: PSF
            An arrays describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """
        return interferometer.SimulatedInterferometer.simulate(
            real_space_image=real_space_image,
            real_space_pixel_scales=self.real_space_pixel_scales,
            exposure_time=self.exposure_time,
            transformer=self.transformer,
            primary_beam=self.primary_beam,
            background_level=self.background_level,
            noise_sigma=self.noise_sigma,
            noise_if_add_noise_false=self.noise_if_add_noise_false,
            noise_seed=self.noise_seed,
        )



from astropy import units
import scipy.signal
from skimage.transform import resize, rescale

import numpy as np

from autoarray.structures import grids, arrays
from autoarray import exc


class Kernel(arrays.AbstractArray):

    # noinspection PyUnusedLocal
    def __new__(cls, array, mask, renormalize=False, *args, **kwargs):
        """ A hyper array with square-pixels.

        Parameters
        ----------
        array: ndarray
            An array representing image (e.g. an image, noise-map, etc.)
        pixel_scales: (float, float)
            The arc-second to pixel conversion factor of each pixel.
        origin : (float, float)
            The arc-second origin of the hyper array's coordinate system.
        """

        #        obj = arrays.Scaled(array_1d=sub_array_1d, mask=mask)

        obj = super(Kernel, cls).__new__(cls=cls, array=array, mask=mask)

        if renormalize:
            obj[:] = np.divide(obj, np.sum(obj))

        return obj

    @classmethod
    def manual_1d(
        cls, array, shape_2d, pixel_scales=None, origin=(0.0, 0.0), renormalize=False
    ):

        array = arrays.Array.manual_1d(
            array=array, shape_2d=shape_2d, pixel_scales=pixel_scales, origin=origin
        )

        return Kernel(array=array, mask=array.mask, renormalize=renormalize)

    @classmethod
    def manual_2d(cls, array, pixel_scales=None, origin=(0.0, 0.0), renormalize=False):

        array = arrays.Array.manual_2d(
            array=array, pixel_scales=pixel_scales, origin=origin
        )

        return Kernel(array=array, mask=array.mask, renormalize=renormalize)

    @classmethod
    def full(
        cls, fill_value, shape_2d, pixel_scales=None, sub_size=1, origin=(0.0, 0.0)
    ):

        if sub_size is not None:
            shape_2d = (shape_2d[0] * sub_size, shape_2d[1] * sub_size)

        return Kernel.manual_2d(
            array=np.full(fill_value=fill_value, shape=shape_2d),
            pixel_scales=pixel_scales,
            origin=origin,
        )

    @classmethod
    def ones(cls, shape_2d, pixel_scales=None, origin=(0.0, 0.0)):
        return Kernel.full(
            fill_value=1.0, shape_2d=shape_2d, pixel_scales=pixel_scales, origin=origin
        )

    @classmethod
    def zeros(cls, shape_2d, pixel_scales=None, origin=(0.0, 0.0)):
        return Kernel.full(
            fill_value=0.0, shape_2d=shape_2d, pixel_scales=pixel_scales, origin=origin
        )

    @classmethod
    def no_blur(cls, pixel_scales=None):

        array = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

        return Kernel.manual_2d(array=array, pixel_scales=pixel_scales)

    @classmethod
    def from_gaussian(
        cls, shape_2d, pixel_scales, sigma, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0
    ):
        """Simulate the Kernel as an elliptical Gaussian profile."""

        grid = grids.Grid.uniform(shape_2d=shape_2d, pixel_scales=pixel_scales)
        grid_shifted = np.subtract(grid, centre)
        grid_radius = np.sqrt(np.sum(grid_shifted ** 2.0, 1))
        theta_coordinate_to_profile = np.arctan2(
            grid_shifted[:, 0], grid_shifted[:, 1]
        ) - np.radians(phi)
        grid_transformed = np.vstack(
            (
                grid_radius * np.sin(theta_coordinate_to_profile),
                grid_radius * np.cos(theta_coordinate_to_profile),
            )
        ).T

        grid_elliptical_radii = np.sqrt(
            np.add(
                np.square(grid_transformed[:, 1]),
                np.square(np.divide(grid_transformed[:, 0], axis_ratio)),
            )
        )

        gaussian = np.multiply(
            np.divide(1.0, sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(grid_elliptical_radii, sigma))),
        )

        return Kernel.manual_1d(
            array=gaussian,
            shape_2d=shape_2d,
            pixel_scales=pixel_scales,
            renormalize=True,
        )

    @classmethod
    def from_as_gaussian_via_alma_fits_header_parameters(
        cls, shape_2d, pixel_scales, y_stddev, x_stddev, theta, centre=(0.0, 0.0)
    ):

        x_stddev = (
            x_stddev * (units.deg).to(units.arcsec) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
        )
        y_stddev = (
            y_stddev * (units.deg).to(units.arcsec) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
        )

        axis_ratio = x_stddev / y_stddev

        return Kernel.from_gaussian(
            shape_2d=shape_2d,
            pixel_scales=pixel_scales,
            sigma=y_stddev,
            axis_ratio=axis_ratio,
            phi=90.0 - theta,
            centre=centre,
        )

    @classmethod
    def from_fits(
        cls, file_path, hdu, pixel_scales=None, origin=(0.0, 0.0), renormalize=False
    ):
        """
        Loads the Kernel from a .fits file.

        Parameters
        ----------
        pixel_scales
        file_path: String
            The path to the file containing the Kernel
        hdu : int
            The HDU the Kernel is stored in the .fits file.
        """

        array = arrays.Array.from_fits(
            file_path=file_path, hdu=hdu, pixel_scales=pixel_scales, origin=origin
        )

        return Kernel(array=array[:], mask=array.mask, renormalize=renormalize)

    def rescaled_with_odd_dimensions_from_rescale_factor(
        self, rescale_factor, renormalize=False
    ):

        kernel_rescaled = rescale(
            self.in_2d,
            rescale_factor,
            anti_aliasing=False,
            mode="constant",
            multichannel=False,
        )

        if kernel_rescaled.shape[0] % 2 == 0 and kernel_rescaled.shape[1] % 2 == 0:
            kernel_rescaled = resize(
                kernel_rescaled,
                output_shape=(
                    kernel_rescaled.shape[0] + 1,
                    kernel_rescaled.shape[1] + 1,
                ),
                anti_aliasing=False,
                mode="constant",
            )
        elif kernel_rescaled.shape[0] % 2 == 0 and kernel_rescaled.shape[1] % 2 != 0:
            kernel_rescaled = resize(
                kernel_rescaled,
                output_shape=(kernel_rescaled.shape[0] + 1, kernel_rescaled.shape[1]),
                anti_aliasing=False,
                mode="constant",
            )
        elif kernel_rescaled.shape[0] % 2 != 0 and kernel_rescaled.shape[1] % 2 == 0:
            kernel_rescaled = resize(
                kernel_rescaled,
                output_shape=(kernel_rescaled.shape[0], kernel_rescaled.shape[1] + 1),
                anti_aliasing=False,
                mode="constant",
            )

        if self.pixel_scales is not None:

            pixel_scale_factors = (
                self.mask.shape[0] / kernel_rescaled.shape[0],
                self.mask.shape[1] / kernel_rescaled.shape[1],
            )

            pixel_scales = (
                self.pixel_scales[0] * pixel_scale_factors[0],
                self.pixel_scales[1] * pixel_scale_factors[1],
            )

        else:

            pixel_scales = None

        return Kernel.manual_2d(
            array=kernel_rescaled, pixel_scales=pixel_scales, renormalize=renormalize
        )

    @property
    def in_2d(self):
        return self.mask.mapping.array_stored_2d_from_sub_array_1d(sub_array_1d=self)

    @property
    def renormalized(self):
        """Renormalize the Kernel such that its data_vector values sum to unity."""
        return Kernel(array=self, mask=self.mask, renormalize=False)

    def convolved_array_from_array(self, array):
        """
        Convolve an array with this Kernel

        Parameters
        ----------
        image : ndarray
            An array representing the image the Kernel is convolved with.

        Returns
        -------
        convolved_image : ndarray
            An array representing the image after convolution.

        Raises
        ------
        KernelException if either Kernel psf dimension is odd
        """
        if self.mask.shape[0] % 2 == 0 or self.mask.shape[1] % 2 == 0:
            raise exc.KernelException("Kernel Kernel must be odd")

        return array.mapping.array_stored_1d_from_array_2d(
            array_2d=scipy.signal.convolve2d(
                array.in_2d_binned, self.in_2d, mode="same"
            )
        )

    def convolved_array_from_array_2d_and_mask(self, array_2d, mask):
        """
        Convolve an array with this Kernel

        Parameters
        ----------
        image : ndarray
            An array representing the image the Kernel is convolved with.

        Returns
        -------
        convolved_image : ndarray
            An array representing the image after convolution.

        Raises
        ------
        KernelException if either Kernel psf dimension is odd
        """

        if self.mask.shape[0] % 2 == 0 or self.mask.shape[1] % 2 == 0:
            raise exc.KernelException("Kernel Kernel must be odd")

        mask_sub_1 = mask.mapping.mask_sub_1

        return mask_sub_1.mapping.array_stored_1d_from_array_2d(
            scipy.signal.convolve2d(array_2d, self.in_2d, mode="same")
        )



import autoarray as aa

import pytest
import numpy as np
import os

test_data_dir = "{}/../test_files/arrays/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestImagingSimulator:
    def test__simulator_grid_is_uniform_grid_with_same_inputs(self):

        grid = aa.grid.uniform(
            shape_2d=(31, 31), pixel_scales=0.05, sub_size=1, origin=(0.1, 0.1)
        )

        simulator = aa.simulator.imaging(
            shape_2d=(31, 31),
            pixel_scales=0.05,
            sub_size=1,
            origin=(0.1, 0.1),
            psf=None,
            exposure_time=20.0,
            background_level=10.0,
        )

        assert (simulator.grid == grid).all()

    def test__constructor_and_specific_instrument_class_methods(self):

        psf = aa.kernel.from_gaussian(shape_2d=(11, 11), sigma=0.1, pixel_scales=0.1)

        simulator = aa.simulator.imaging(
            shape_2d=(51, 51),
            pixel_scales=0.1,
            sub_size=1,
            psf=psf,
            exposure_time=20.0,
            background_level=10.0,
        )

        assert simulator.shape_2d == (51, 51)
        assert simulator.pixel_scales == (0.1, 0.1)
        assert simulator.psf == psf
        assert simulator.exposure_time == 20.0
        assert simulator.background_level == 10.0

        lsst = aa.simulator.imaging.lsst()

        lsst_psf = aa.kernel.from_gaussian(
            shape_2d=(31, 31), sigma=0.5, pixel_scales=0.2
        )

        assert lsst.shape_2d == (101, 101)
        assert lsst.pixel_scales == (0.2, 0.2)
        assert lsst.psf == lsst_psf
        assert lsst.exposure_time == 100.0
        assert lsst.background_level == 1.0

        euclid = aa.simulator.imaging.euclid()

        euclid_psf = aa.kernel.from_gaussian(
            shape_2d=(31, 31), sigma=0.1, pixel_scales=0.1
        )

        assert euclid.shape_2d == (151, 151)
        assert euclid.pixel_scales == (0.1, 0.1)
        assert euclid.psf == euclid_psf
        assert euclid.exposure_time == 565.0
        assert euclid.background_level == 1.0

        hst = aa.simulator.imaging.hst()

        hst_psf = aa.kernel.from_gaussian(
            shape_2d=(31, 31), sigma=0.05, pixel_scales=0.05
        )

        assert hst.shape_2d == (251, 251)
        assert hst.pixel_scales == (0.05, 0.05)
        assert hst.psf == hst_psf
        assert hst.exposure_time == 2000.0
        assert hst.background_level == 1.0

        hst_up_sampled = aa.simulator.imaging.hst_up_sampled()

        hst_up_sampled_psf = aa.kernel.from_gaussian(
            shape_2d=(31, 31), sigma=0.05, pixel_scales=0.03
        )

        assert hst_up_sampled.shape_2d == (401, 401)
        assert hst_up_sampled.pixel_scales == (0.03, 0.03)
        assert hst_up_sampled.psf == hst_up_sampled_psf
        assert hst_up_sampled.exposure_time == 2000.0
        assert hst_up_sampled.background_level == 1.0

        adaptive_optics = aa.simulator.imaging.keck_adaptive_optics()

        adaptive_optics_psf = aa.kernel.from_gaussian(
            shape_2d=(31, 31), sigma=0.025, pixel_scales=0.01
        )

        assert adaptive_optics.shape_2d == (751, 751)
        assert adaptive_optics.pixel_scales == (0.01, 0.01)
        assert adaptive_optics.psf == adaptive_optics_psf
        assert adaptive_optics.exposure_time == 1000.0
        assert adaptive_optics.background_level == 1.0

    def test__from_image_same_as_manual_image_input(self):

        psf = aa.kernel.manual_2d(
            array=np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]]),
            pixel_scales=1.0,
        )

        image = aa.array.manual_2d([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])

        simulator = aa.simulator.imaging(
            shape_2d=(20, 20),
            pixel_scales=0.05,
            sub_size=1,
            psf=psf,
            exposure_time=10000.0,
            background_level=100.0,
            add_noise=True,
            noise_seed=1,
        )

        imaging_simulated = simulator.from_image(image=image)

        imaging_manual = aa.imaging.simulate(
            image=image,
            exposure_time=10000.0,
            psf=psf,
            background_level=100.0,
            add_noise=True,
            noise_seed=1,
        )

        assert (imaging_simulated.image.in_2d == imaging_manual.image.in_2d).all()
        assert (imaging_simulated.psf == imaging_manual.psf).all()
        assert (imaging_simulated.noise_map == imaging_manual.noise_map).all()
        assert (
            imaging_simulated.background_sky_map == imaging_manual.background_sky_map
        ).all()
        assert (
            imaging_simulated.exposure_time_map == imaging_manual.exposure_time_map
        ).all()


class TestInterferometerSimulator:
    def test__simulator_grid_is_uniform_grid_with_same_inputs(self):

        grid = aa.grid.uniform(
            shape_2d=(31, 31), pixel_scales=0.05, sub_size=1, origin=(0.1, 0.1)
        )

        simulator = aa.simulator.interferometer(
            real_space_shape_2d=(31, 31),
            real_space_pixel_scales=0.05,
            uv_wavelengths=np.ones((7, 2)),
            sub_size=1,
            origin=(0.1, 0.1),
            exposure_time=20.0,
            background_level=10.0,
        )

        assert (simulator.grid == grid).all()

    def test__constructor_and_specific_instrument_class_methods(self):

        sma = aa.simulator.interferometer.sma()

        uv_wavelengths_path = "{}/dataset/sma_uv_wavelengths.fits".format(
            os.path.dirname(os.path.realpath(__file__))
        )

        sma_uv_wavelengths = aa.util.array.numpy_array_1d_from_fits(
            file_path=uv_wavelengths_path, hdu=0
        )

        assert sma.real_space_shape_2d == (151, 151)
        assert sma.real_space_pixel_scales == (0.05, 0.05)
        assert (sma.uv_wavelengths == sma_uv_wavelengths).all()
        assert sma.uv_wavelengths[0] == pytest.approx(
            [184584.953125, -16373.30566406], 1.0e-4
        )
        assert sma.exposure_time == 100.0
        assert sma.background_level == 1.0

    def test__from_real_space_image_same_as_manual_image_input(self):

        primary_beam = aa.kernel.manual_2d(
            array=np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]]),
            pixel_scales=1.0,
        )

        real_space_image = aa.array.manual_2d([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])

        simulator = aa.simulator.interferometer(
            real_space_shape_2d=(2, 3),
            real_space_pixel_scales=0.05,
            uv_wavelengths=np.ones((7, 2)),
            sub_size=1,
            primary_beam=primary_beam,
            exposure_time=10000.0,
            background_level=100.0,
            noise_sigma=0.1,
            noise_seed=1,
        )

        interferometer_simulated = simulator.from_real_space_image(
            real_space_image=real_space_image
        )

        interferometer_manual = aa.interferometer.simulate(
            real_space_image=real_space_image,
            real_space_pixel_scales=0.05,
            transformer=simulator.transformer,
            exposure_time=10000.0,
            primary_beam=primary_beam,
            background_level=100.0,
            noise_sigma=0.1,
            noise_seed=1,
        )

        assert (
            interferometer_simulated.visibilities == interferometer_manual.visibilities
        ).all()
        assert (
            interferometer_simulated.noise_map == interferometer_manual.noise_map
        ).all()
        assert (
            interferometer_simulated.primary_beam.in_2d
            == interferometer_manual.primary_beam.in_2d
        ).all()



import logging
import os

import numpy as np

import autoarray as aa
from autoarray.dataset import abstract_dataset

logger = logging.getLogger(__name__)

test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestSignalToNoise:
    def test__image_and_noise_are_values__signal_to_noise_is_ratio_of_each(self):
        array = aa.array.manual_2d([[1.0, 2.0], [3.0, 4.0]])
        noise_map = aa.array.manual_2d([[10.0, 10.0], [30.0, 4.0]])

        dataset = abstract_dataset.AbstractDataset(data=array, noise_map=noise_map)

        assert (
            dataset.signal_to_noise_map.in_2d == np.array([[0.1, 0.2], [0.1, 1.0]])
        ).all()
        assert dataset.signal_to_noise_max == 1.0

    def test__same_as_above__but_image_has_negative_values__replaced_with_zeros(self):
        array = aa.array.manual_2d([[-1.0, 2.0], [3.0, -4.0]])

        noise_map = aa.array.manual_2d([[10.0, 10.0], [30.0, 4.0]])

        dataset = abstract_dataset.AbstractDataset(data=array, noise_map=noise_map)

        assert (
            dataset.signal_to_noise_map.in_2d == np.array([[0.0, 0.2], [0.1, 0.0]])
        ).all()
        assert dataset.signal_to_noise_max == 0.2


class TestAbsoluteSignalToNoise:
    def test__image_and_noise_are_values__signal_to_noise_is_absolute_image_value_over_noise(
        self
    ):
        array = aa.array.manual_2d([[-1.0, 2.0], [3.0, -4.0]])

        noise_map = aa.array.manual_2d([[10.0, 10.0], [30.0, 4.0]])

        dataset = abstract_dataset.AbstractDataset(data=array, noise_map=noise_map)

        assert (
            dataset.absolute_signal_to_noise_map.in_2d
            == np.array([[0.1, 0.2], [0.1, 1.0]])
        ).all()
        assert dataset.absolute_signal_to_noise_max == 1.0


class TestPotentialChiSquaredMap:
    def test__image_and_noise_are_values__signal_to_noise_is_absolute_image_value_over_noise(
        self
    ):
        array = aa.array.manual_2d([[-1.0, 2.0], [3.0, -4.0]])
        noise_map = aa.array.manual_2d([[10.0, 10.0], [30.0, 4.0]])

        dataset = abstract_dataset.AbstractDataset(data=array, noise_map=noise_map)

        assert (
            dataset.potential_chi_squared_map.in_2d
            == np.array([[0.1 ** 2.0, 0.2 ** 2.0], [0.1 ** 2.0, 1.0 ** 2.0]])
        ).all()
        assert dataset.potential_chi_squared_max == 1.0


class TestExposureTimeMap:
    def test__from_background_noise_map__covnerts_to_exposure_times(self):
        background_noise_map = aa.array.manual_2d([[1.0, 4.0, 8.0], [1.0, 4.0, 8.0]])

        exposure_time_map = abstract_dataset.ExposureTimeMap.from_exposure_time_and_inverse_noise_map(
            exposure_time=1.0, inverse_noise_map=background_noise_map
        )

        assert (
            exposure_time_map.in_2d == np.array([[0.125, 0.5, 1.0], [0.125, 0.5, 1.0]])
        ).all()

        exposure_time_map = abstract_dataset.ExposureTimeMap.from_exposure_time_and_inverse_noise_map(
            exposure_time=3.0, inverse_noise_map=background_noise_map
        )

        assert (
            exposure_time_map.in_2d == np.array([[0.375, 1.5, 3.0], [0.375, 1.5, 3.0]])
        ).all()



import numpy as np
import autoarray as aa


class TestNoiseMapFromWeightMap:
    def test__weight_map_no_zeros__uses_1_over_sqrt_value(self):
        weight_map = aa.array.manual_2d([[1.0, 4.0, 16.0], [1.0, 4.0, 16.0]])

        noise_map = aa.data_converter.noise_map_from_weight_map(
            weight_map=weight_map.in_2d
        )

        assert (noise_map.in_2d == np.array([[1.0, 0.5, 0.25], [1.0, 0.5, 0.25]])).all()

    def test__weight_map_no_zeros__zeros_set_to_10000000(self):

        weight_map = aa.array.manual_2d([[1.0, 4.0, 0.0], [1.0, 4.0, 16.0]])

        noise_map = aa.data_converter.noise_map_from_weight_map(weight_map=weight_map)

        assert (
            noise_map.in_2d == np.array([[1.0, 0.5, 1.0e8], [1.0, 0.5, 0.25]])
        ).all()


class TestFromInverseAbstractNoiseMap:
    def test__inverse_noise_map_no_zeros__uses_1_over_value(self):
        inverse_noise_map = aa.array.manual_2d([[1.0, 4.0, 16.0], [1.0, 4.0, 16.0]])

        noise_map = aa.data_converter.noise_map_from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map
        )

        assert (
            noise_map.in_2d == np.array([[1.0, 0.25, 0.0625], [1.0, 0.25, 0.0625]])
        ).all()


class TestFromImageAndBackgroundNoiseMap:
    def test__image_all_1s__bg_noise_all_1s__exposure_time_all_1s__noise_map_all_sqrt_2s(
        self
    ):
        imaging = aa.array.ones(shape_2d=(2, 2))
        background_noise_map = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.ones(shape_2d=(2, 2))

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            gain=1.0,
            exposure_time_map=exposure_time_map,
        )

        assert (
            noise_map.in_2d
            == np.array([[np.sqrt(2.0), np.sqrt(2.0)], [np.sqrt(2.0), np.sqrt(2.0)]])
        ).all()

    def test__image_all_2s__bg_noise_all_1s__exposure_time_all_1s__noise_map_all_sqrt_3s(
        self
    ):
        imaging = aa.array.full(fill_value=2.0, shape_2d=(2, 2))
        background_noise_map = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.ones(shape_2d=(2, 2))

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            gain=1.0,
            exposure_time_map=exposure_time_map,
        )

        assert (
            noise_map.in_2d
            == np.array([[np.sqrt(3.0), np.sqrt(3.0)], [np.sqrt(3.0), np.sqrt(3.0)]])
        ).all()

    def test__image_all_1s__bg_noise_all_2s__exposure_time_all_1s__noise_map_all_sqrt_5s(
        self
    ):
        imaging = aa.array.ones(shape_2d=(2, 2))
        background_noise_map = aa.array.full(fill_value=2.0, shape_2d=(2, 2))
        exposure_time_map = aa.array.ones(shape_2d=(2, 2))

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            gain=1.0,
            exposure_time_map=exposure_time_map,
        )

        assert (
            noise_map.in_2d
            == np.array([[np.sqrt(5.0), np.sqrt(5.0)], [np.sqrt(5.0), np.sqrt(5.0)]])
        ).all()

    def test__image_all_1s__bg_noise_all_1s__exposure_time_all_2s__noise_map_all_sqrt_6s_over_2(
        self
    ):
        imaging = aa.array.ones(shape_2d=(2, 2))
        background_noise_map = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.full(fill_value=2.0, shape_2d=(2, 2))

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            gain=1.0,
            exposure_time_map=exposure_time_map,
        )

        assert (
            noise_map.in_2d
            == np.array(
                [
                    [np.sqrt(6.0) / 2.0, np.sqrt(6.0) / 2.0],
                    [np.sqrt(6.0) / 2.0, np.sqrt(6.0) / 2.0],
                ]
            )
        ).all()

    def test__image_all_negative_2s__bg_noise_all_1s__exposure_time_all_1s__noise_map_all_1s(
        self
    ):
        imaging = aa.array.full(fill_value=-2.0, shape_2d=(2, 2))
        background_noise_map = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.ones(shape_2d=(2, 2))

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            gain=1.0,
            exposure_time_map=exposure_time_map,
        )

        assert (noise_map.in_2d == np.array([[1.0, 1.0], [1.0, 1.0]])).all()

    def test__same_as_above__use_different_values_in_different_array_elemets(self):
        imaging = aa.array.manual_2d([[1.0, 2.0], [2.0, 3.0]])
        background_noise_map = aa.array.manual_2d([[1.0, 1.0], [2.0, 3.0]])
        exposure_time_map = aa.array.manual_2d([[4.0, 3.0], [2.0, 1.0]])

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            gain=1.0,
            exposure_time_map=exposure_time_map,
        )

        assert (
            noise_map.in_2d
            == np.array(
                [
                    [np.sqrt(20.0) / 4.0, np.sqrt(15.0) / 3.0],
                    [np.sqrt(20.0) / 2.0, np.sqrt(12.0)],
                ]
            )
        ).all()

    def test__convert_from_electrons__image_all_1s__bg_noise_all_1s__exposure_time_all_1s__noise_map_all_sqrt_2s(
        self
    ):
        imaging = aa.array.ones(shape_2d=(2, 2))
        background_noise_map = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.ones(shape_2d=(2, 2))

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            exposure_time_map=exposure_time_map,
            gain=2.0,
            convert_from_electrons=True,
        )

        assert (
            noise_map.in_2d
            == np.array([[np.sqrt(2.0), np.sqrt(2.0)], [np.sqrt(2.0), np.sqrt(2.0)]])
        ).all()

    def test__convert_from_electrons__image_all_negative_2s__bg_noise_all_1s__exposure_time_all_10s__noise_map_all_1s(
        self
    ):
        imaging = aa.array.full(fill_value=-2.0, shape_2d=(2, 2))
        background_noise_map = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.full(fill_value=10.0, shape_2d=(2, 2))

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            exposure_time_map=exposure_time_map,
            gain=1.0,
            convert_from_electrons=True,
        )

        assert (noise_map.in_2d == np.array([[1.0, 1.0], [1.0, 1.0]])).all()

    def test__convert_from_electrons__same_as_above__use_different_values_in_different_array_elemets(
        self
    ):
        imaging = aa.array.manual_2d([[1.0, 2.0], [2.0, 3.0]])
        background_noise_map = aa.array.manual_2d([[1.0, 1.0], [2.0, 3.0]])
        exposure_time_map = aa.array.manual_2d([[10.0, 11.0], [12.0, 13.0]])

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            exposure_time_map=exposure_time_map,
            gain=4.0,
            convert_from_electrons=True,
        )

        assert (
            noise_map.in_2d
            == np.array([[np.sqrt(2.0), np.sqrt(3.0)], [np.sqrt(6.0), np.sqrt(12.0)]])
        ).all()

    def test__convert_from_adus__same_as_above__gain_is_1__same_values(self):
        imaging = aa.array.manual_2d([[1.0, 2.0], [2.0, 3.0]])
        background_noise_map = aa.array.manual_2d([[1.0, 1.0], [2.0, 3.0]])
        exposure_time_map = aa.array.manual_2d([[10.0, 11.0], [12.0, 13.0]])

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            exposure_time_map=exposure_time_map,
            gain=1.0,
            convert_from_adus=True,
        )

        assert (
            noise_map.in_2d
            == np.array([[np.sqrt(2.0), np.sqrt(3.0)], [np.sqrt(6.0), np.sqrt(12.0)]])
        ).all()

    def test__convert_from_adus__same_as_above__gain_is_2__values_change(self):
        imaging = aa.array.manual_2d([[1.0, 2.0], [2.0, 3.0]])
        background_noise_map = aa.array.manual_2d([[1.0, 1.0], [2.0, 3.0]])
        exposure_time_map = aa.array.manual_2d([[10.0, 11.0], [12.0, 13.0]])

        noise_map = aa.data_converter.noise_map_from_image_and_background_noise_map(
            image=imaging,
            background_noise_map=background_noise_map,
            exposure_time_map=exposure_time_map,
            gain=2.0,
            convert_from_adus=True,
        )

        assert (
            noise_map.in_2d
            == np.array(
                [
                    [np.sqrt(6.0) / 2.0, np.sqrt(8.0) / 2.0],
                    [np.sqrt(20.0) / 2.0, np.sqrt(42.0) / 2.0],
                ]
            )
        ).all()


class TestFromImageAndExposureTimeMap:
    def test__image_all_1s__exposure_time_all_1s__noise_map_all_1s(self):
        imaging = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.ones(shape_2d=(2, 2))

        poisson_noise_map = aa.data_converter.poisson_noise_map_from_image_and_exposure_time_map(
            image=imaging, exposure_time_map=exposure_time_map, gain=1.0
        )

        assert (poisson_noise_map.in_2d == np.array([[1.0, 1.0], [1.0, 1.0]])).all()

    def test__image_all_2s_and_3s__exposure_time_all_1s__noise_map_all_sqrt_2s_and_3s(
        self
    ):
        imaging = aa.array.manual_2d([[2.0, 2.0], [3.0, 3.0]])
        exposure_time_map = aa.array.ones(shape_2d=(2, 2))

        poisson_noise_map = aa.data_converter.poisson_noise_map_from_image_and_exposure_time_map(
            image=imaging, exposure_time_map=exposure_time_map, gain=1.0
        )

        assert (
            poisson_noise_map.in_2d
            == np.array([[np.sqrt(2.0), np.sqrt(2.0)], [np.sqrt(3.0), np.sqrt(3.0)]])
        ).all()

    def test__image_all_1s__exposure_time_all__2s_and_3s__noise_map_all_sqrt_2s_and_3s(
        self
    ):
        imaging = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.manual_2d([[2.0, 2.0], [3.0, 3.0]])

        poisson_noise_map = aa.data_converter.poisson_noise_map_from_image_and_exposure_time_map(
            image=imaging, exposure_time_map=exposure_time_map, gain=1.0
        )

        assert (
            poisson_noise_map.in_2d
            == np.array(
                [
                    [np.sqrt(2.0) / 2.0, np.sqrt(2.0) / 2.0],
                    [np.sqrt(3.0) / 3.0, np.sqrt(3.0) / 3.0],
                ]
            )
        ).all()

    def test__image_all_1s__exposure_time_all_1s__noise_map_all_1s__gain_is_2__ignores_gain(
        self
    ):
        imaging = aa.array.ones(shape_2d=(2, 2))
        exposure_time_map = aa.array.ones(shape_2d=(2, 2))

        poisson_noise_map = aa.data_converter.poisson_noise_map_from_image_and_exposure_time_map(
            image=imaging, exposure_time_map=exposure_time_map, gain=2.0
        )

        assert (
            poisson_noise_map.in_2d
            == np.array([[np.sqrt(1.0), np.sqrt(1.0)], [np.sqrt(1.0), np.sqrt(1.0)]])
        ).all()

    def test__convert_from_electrons_is_true__image_already_in_counts_so_exposure_time_ignored(
        self
    ):
        imaging = aa.array.manual_2d([[2.0, 2.0], [3.0, 3.0]])
        exposure_time_map = aa.array.full(fill_value=10.0, shape_2d=(2, 2))

        poisson_noise_map = aa.data_converter.poisson_noise_map_from_image_and_exposure_time_map(
            image=imaging,
            exposure_time_map=exposure_time_map,
            gain=4.0,
            convert_from_electrons=True,
        )

        assert (
            poisson_noise_map.in_2d
            == np.array([[np.sqrt(2.0), np.sqrt(2.0)], [np.sqrt(3.0), np.sqrt(3.0)]])
        ).all()

    def test__same_as_above__convert_from_adus__includes_gain_multiplication(self):
        imaging = aa.array.manual_2d([[2.0, 2.0], [3.0, 3.0]])
        exposure_time_map = aa.array.full(fill_value=10.0, shape_2d=(2, 2))

        poisson_noise_map = aa.data_converter.poisson_noise_map_from_image_and_exposure_time_map(
            image=imaging,
            exposure_time_map=exposure_time_map,
            gain=2.0,
            convert_from_adus=True,
        )

        assert (
            poisson_noise_map.in_2d
            == np.array(
                [
                    [np.sqrt(2.0 * 2.0) / 2.0, np.sqrt(2.0 * 2.0) / 2.0],
                    [np.sqrt(2.0 * 3.0) / 2.0, np.sqrt(2.0 * 3.0) / 2.0],
                ]
            )
        ).all()



import logging

import numpy as np
from scipy.stats import norm

from autoarray import exc
from autoarray.dataset import abstract_dataset, data_converter
from autoarray.mask import mask as msk
from autoarray.masked import masked_structures
from autoarray.structures import kernel, arrays

logger = logging.getLogger(__name__)


class AbstractImagingDataSet(abstract_dataset.AbstractDataset):
    @property
    def image(self):
        return self.data

    @property
    def pixel_scales(self):
        return self.data.pixel_scales

    @property
    def pixel_scale(self):
        return self.data.pixel_scale

    @property
    def shape(self):
        return self.image.shape_2d

    @property
    def shape_2d(self):
        return self.image.shape_2d

    def binned_from_bin_up_factor(self, bin_up_factor):

        image = self.image.binned_from_bin_up_factor(
            bin_up_factor=bin_up_factor, method="mean"
        )
        psf = self.psf.rescaled_with_odd_dimensions_from_rescale_factor(
            rescale_factor=1.0 / bin_up_factor, renormalize=False
        )
        noise_map = (
            self.noise_map.binned_from_bin_up_factor(
                bin_up_factor=bin_up_factor, method="quadrature"
            )
            if self.noise_map is not None
            else None
        )

        background_noise_map = (
            self.background_noise_map.binned_from_bin_up_factor(
                bin_up_factor=bin_up_factor, method="quadrature"
            )
            if self.background_noise_map is not None
            else None
        )

        poisson_noise_map = (
            self.poisson_noise_map.binned_from_bin_up_factor(
                bin_up_factor=bin_up_factor, method="quadrature"
            )
            if self.poisson_noise_map is not None
            else None
        )

        exposure_time_map = (
            self.exposure_time_map.binned_from_bin_up_factor(
                bin_up_factor=bin_up_factor, method="sum"
            )
            if self.exposure_time_map is not None
            else None
        )

        background_sky_map = (
            self.background_sky_map.binned_from_bin_up_factor(
                bin_up_factor=bin_up_factor, method="mean"
            )
            if self.background_sky_map is not None
            else None
        )

        return Imaging(
            image=image,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            name=self.name,
        )

    def resized_from_new_shape(self, new_shape):

        image = self.image.resized_from_new_shape(new_shape=new_shape)

        noise_map = (
            self.noise_map.resized_from_new_shape(new_shape=new_shape)
            if self.noise_map is not None
            else None
        )

        background_noise_map = (
            self.background_noise_map.resized_from_new_shape(new_shape=new_shape)
            if self.background_noise_map is not None
            else None
        )

        poisson_noise_map = (
            self.poisson_noise_map.resized_from_new_shape(new_shape=new_shape)
            if self.poisson_noise_map is not None
            else None
        )

        exposure_time_map = (
            self.exposure_time_map.resized_from_new_shape(new_shape=new_shape)
            if self.exposure_time_map is not None
            else None
        )

        background_sky_map = (
            self.background_sky_map.resized_from_new_shape(new_shape=new_shape)
            if self.background_sky_map is not None
            else None
        )

        return Imaging(
            image=image,
            psf=self.psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            name=self.name,
        )

    def resized_psf_from_new_shape(self, new_shape):
        psf = self.psf.resized_from_new_shape(new_shape=new_shape)
        return Imaging(
            image=self.image,
            psf=psf,
            noise_map=self.noise_map,
            background_noise_map=self.background_noise_map,
            poisson_noise_map=self.poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=self.background_sky_map,
            name=self.name,
        )

    def modified_image_from_image(self, image):

        return Imaging(
            image=image,
            psf=self.psf,
            noise_map=self.noise_map,
            background_noise_map=self.background_noise_map,
            poisson_noise_map=self.poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=self.background_sky_map,
            name=self.name,
        )

    def add_poisson_noise_to_data(self, seed=-1):

        image_with_sky = self.image + self.background_sky_map

        image_with_sky_and_noise = image_with_sky + generate_poisson_noise(
            image=image_with_sky, exposure_time_map=self.exposure_time_map, seed=seed
        )

        image_with_noise = image_with_sky_and_noise - self.background_sky_map

        return Imaging(
            image=image_with_noise,
            psf=self.psf,
            noise_map=self.noise_map,
            background_noise_map=self.background_noise_map,
            poisson_noise_map=self.poisson_noise_map,
            name=self.name,
        )

    def data_in_electrons(self):

        image = self.array_from_counts_to_electrons_per_second(array=self.image)
        noise_map = self.array_from_counts_to_electrons_per_second(array=self.noise_map)
        background_noise_map = self.array_from_counts_to_electrons_per_second(
            array=self.background_noise_map
        )
        poisson_noise_map = self.array_from_counts_to_electrons_per_second(
            array=self.poisson_noise_map
        )
        background_sky_map = self.array_from_counts_to_electrons_per_second(
            array=self.background_sky_map
        )

        return Imaging(
            image=image,
            psf=self.psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=background_sky_map,
            name=self.name,
        )

    def data_in_adus_from_gain(self, gain):

        image = self.array_from_adus_to_electrons_per_second(
            array=self.image, gain=gain
        )
        noise_map = self.array_from_adus_to_electrons_per_second(
            array=self.noise_map, gain=gain
        )
        background_noise_map = self.array_from_adus_to_electrons_per_second(
            array=self.background_noise_map, gain=gain
        )
        poisson_noise_map = self.array_from_adus_to_electrons_per_second(
            array=self.poisson_noise_map, gain=gain
        )
        background_sky_map = self.array_from_adus_to_electrons_per_second(
            array=self.background_sky_map, gain=gain
        )

        return Imaging(
            image=image,
            psf=self.psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=background_sky_map,
            name=self.name,
        )

    def signal_to_noise_limited_from_signal_to_noise_limit(self, signal_to_noise_limit):

        noise_map_limit = np.where(
            self.signal_to_noise_map > signal_to_noise_limit,
            np.abs(self.image) / signal_to_noise_limit,
            self.noise_map,
        )

        noise_map_limit = masked_structures.MaskedArray.manual_1d(
            array=noise_map_limit, mask=self.image.mask
        )

        return Imaging(
            image=self.image,
            psf=self.psf,
            noise_map=noise_map_limit,
            background_noise_map=self.background_noise_map,
            poisson_noise_map=self.poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=self.background_sky_map,
            name=self.name,
        )

    @property
    def background_noise_map_counts(self):
        """ The background noise_maps mappers in unit_label of counts."""
        return self.array_from_electrons_per_second_to_counts(self.background_noise_map)

    @property
    def estimated_noise_map_counts(self):
        """ The estimated noise_maps mappers of the image (using its background noise_maps mappers and image values
        in counts) in counts.
        """
        return np.sqrt(
            (np.abs(self.image_counts) + np.square(self.background_noise_map_counts))
        )

    @property
    def estimated_noise_map(self):
        """ The estimated noise_maps mappers of the image (using its background noise_maps mappers and image values
        in counts) in electrons per second.
        """
        return self.array_from_counts_to_electrons_per_second(
            self.estimated_noise_map_counts
        )

    def background_noise_from_edges(self, no_edges):
        """Estimate the background signal_to_noise_ratio by binning data_to_image located at the edge(s) of an image
        into a histogram and fitting a Gaussian profiles to this histogram. The standard deviation (sigma) of this
        Gaussian gives a signal_to_noise_ratio estimate.

        Parameters
        ----------
        no_edges : int
            Number of edges used to estimate the background signal_to_noise_ratio.

        """

        edges = []

        for edge_no in range(no_edges):
            top_edge = self.image.in_2d[
                edge_no, edge_no : self.image.shape_2d[1] - edge_no
            ]
            bottom_edge = self.image.in_2d[
                self.image.shape_2d[0] - 1 - edge_no,
                edge_no : self.image.shape_2d[1] - edge_no,
            ]
            left_edge = self.image.in_2d[
                edge_no + 1 : self.image.shape_2d[0] - 1 - edge_no, edge_no
            ]
            right_edge = self.image.in_2d[
                edge_no + 1 : self.image.shape_2d[0] - 1 - edge_no,
                self.image.shape_2d[1] - 1 - edge_no,
            ]

            edges = np.concatenate(
                (edges, top_edge, bottom_edge, right_edge, left_edge)
            )

        return norm.fit(edges)[1]

    def output_to_fits(
        self,
        image_path,
        psf_path=None,
        noise_map_path=None,
        background_noise_map_path=None,
        poisson_noise_map_path=None,
        exposure_time_map_path=None,
        background_sky_map_path=None,
        overwrite=False,
    ):
        self.image.output_to_fits(file_path=image_path, overwrite=overwrite)

        if self.psf is not None and psf_path is not None:
            self.psf.output_to_fits(file_path=psf_path, overwrite=overwrite)

        if self.noise_map is not None and noise_map_path is not None:
            self.noise_map.output_to_fits(file_path=noise_map_path, overwrite=overwrite)

        if (
            self.background_noise_map is not None
            and background_noise_map_path is not None
        ):
            self.background_noise_map.output_to_fits(
                file_path=background_noise_map_path, overwrite=overwrite
            )

        if self.poisson_noise_map is not None and poisson_noise_map_path is not None:
            self.poisson_noise_map.output_to_fits(
                file_path=poisson_noise_map_path, overwrite=overwrite
            )

        if self.exposure_time_map is not None and exposure_time_map_path is not None:
            self.exposure_time_map.output_to_fits(
                file_path=exposure_time_map_path, overwrite=overwrite
            )

        if self.background_sky_map is not None and background_sky_map_path is not None:
            self.background_sky_map.output_to_fits(
                file_path=background_sky_map_path, overwrite=overwrite
            )


class Imaging(AbstractImagingDataSet):
    def __init__(
        self,
        image,
        noise_map,
        psf=None,
        background_noise_map=None,
        poisson_noise_map=None,
        exposure_time_map=None,
        background_sky_map=None,
        name=None,
        **kwargs
    ):
        """A collection of 2D imaging dataset(an image, noise-map, psf, etc.)

        Parameters
        ----------
        image : aa.Array
            The array of the image data_type, in unit_label of electrons per second.
        psf : PSF
            An array describing the PSF kernel of the image.
        noise_map : NoiseMap | float | ndarray
            An array describing the RMS standard deviation error in each pixel, preferably in unit_label of electrons per
            second.
        background_noise_map : NoiseMap
            An array describing the RMS standard deviation error in each pixel due to the background sky noise_map,
            preferably in unit_label of electrons per second.
        poisson_noise_map : NoiseMap
            An array describing the RMS standard deviation error in each pixel due to the Poisson counts of the source,
            preferably in unit_label of electrons per second.
        exposure_time_map : aa.Array
            An array describing the effective exposure time in each imaging pixel.
        background_sky_map : aa.Scaled
            An array describing the background sky.
        """

        super(Imaging, self).__init__(
            data=image, noise_map=noise_map, exposure_time_map=exposure_time_map
        )

        self.psf = psf
        self.__name = name
        self.background_noise_map = background_noise_map
        self.poisson_noise_map = poisson_noise_map
        self.background_sky_map = background_sky_map

    @property
    def name(self) -> str:
        return self.__name

    @classmethod
    def manual(
        cls,
        image,
        noise_map,
        psf=None,
        background_noise_map=None,
        poisson_noise_map=None,
        exposure_time_map=None,
        background_sky_map=None,
    ):
        return Imaging(
            image=image,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
        )

    @classmethod
    def from_fits(
        cls,
        image_path,
        pixel_scales=None,
        image_hdu=0,
        resized_imaging_shape=None,
        noise_map_path=None,
        noise_map_hdu=0,
        noise_map_from_image_and_background_noise_map=False,
        noise_map_non_constant=False,
        psf_path=None,
        psf_hdu=0,
        resized_psf_shape=None,
        renormalize_psf=True,
        convert_noise_map_from_weight_map=False,
        convert_noise_map_from_inverse_noise_map=False,
        background_noise_map_path=None,
        background_noise_map_hdu=0,
        convert_background_noise_map_from_weight_map=False,
        convert_background_noise_map_from_inverse_noise_map=False,
        poisson_noise_map_path=None,
        poisson_noise_map_hdu=0,
        poisson_noise_map_from_image=False,
        convert_poisson_noise_map_from_weight_map=False,
        convert_poisson_noise_map_from_inverse_noise_map=False,
        exposure_time_map_path=None,
        exposure_time_map_hdu=0,
        exposure_time_map_from_single_value=None,
        exposure_time_map_from_inverse_noise_map=False,
        background_sky_map_path=None,
        background_sky_map_hdu=0,
        convert_from_electrons=False,
        gain=None,
        convert_from_adus=False,
        lens_name=None,
    ):
        """Factory for loading the imaging data_type from .fits files, as well as computing properties like the noise-map,
        exposure-time map, etc. from the imaging-data.

        This factory also includes a number of routines for converting the imaging-data from unit_label not supported by PyAutoLens \
        (e.g. adus, electrons) to electrons per second.

        Parameters
        ----------
        lens_name
        image_path : str
            The path to the image .fits file containing the image (e.g. '/path/to/image.fits')
        pixel_scales : float
            The size of each pixel in arc seconds.
        image_hdu : int
            The hdu the image is contained in the .fits file specified by *image_path*.
        image_hdu : int
            The hdu the image is contained in the .fits file that *image_path* points too.
        resized_imaging_shape : (int, int) | None
            If input, the imaging structures that are image sized, e.g. the image, noise-maps) are resized to these dimensions.
        resized_imaging_origin_pixels : (int, int) | None
            If the imaging structures are resized, this defines a new origin (in pixels) around which recentering occurs.
        resized_imaging_origin_arcsec : (float, float) | None
            If the imaging structures are resized, this defines a new origin (in arc-seconds) around which recentering occurs.
        psf_path : str
            The path to the psf .fits file containing the psf (e.g. '/path/to/psf.fits')
        psf_hdu : int
            The hdu the psf is contained in the .fits file specified by *psf_path*.
        resized_psf_shape : (int, int) | None
            If input, the psf is resized to these dimensions.
        renormalize_psf : bool
            If True, the PSF is renoralized such that all elements sum to 1.0.
        noise_map_path : str
            The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')
        noise_map_hdu : int
            The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
        noise_map_from_image_and_background_noise_map : bool
            If True, the noise-map is computed from the observed image and background noise-map \
            (see NoiseMap.from_image_and_background_noise_map).
        convert_noise_map_from_weight_map : bool
            If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
            *NoiseMap.from_weight_map).
        convert_noise_map_from_inverse_noise_map : bool
            If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
            *NoiseMap.from_inverse_noise_map).
        background_noise_map_path : str
            The path to the background_noise_map .fits file containing the background noise-map \
            (e.g. '/path/to/background_noise_map.fits')
        background_noise_map_hdu : int
            The hdu the background_noise_map is contained in the .fits file specified by *background_noise_map_path*.
        convert_background_noise_map_from_weight_map : bool
            If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
            *NoiseMap.from_weight_map).
        convert_background_noise_map_from_inverse_noise_map : bool
            If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
            noise-map (see *NoiseMap.from_inverse_noise_map).
        poisson_noise_map_path : str
            The path to the poisson_noise_map .fits file containing the Poisson noise-map \
             (e.g. '/path/to/poisson_noise_map.fits')
        poisson_noise_map_hdu : int
            The hdu the poisson_noise_map is contained in the .fits file specified by *poisson_noise_map_path*.
        poisson_noise_map_from_image : bool
            If True, the Poisson noise-map is estimated using the image.
        convert_poisson_noise_map_from_weight_map : bool
            If True, the Poisson noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
            *NoiseMap.from_weight_map).
        convert_poisson_noise_map_from_inverse_noise_map : bool
            If True, the Poisson noise-map loaded from the .fits file is converted from an inverse noise-map to a \
            noise-map (see *NoiseMap.from_inverse_noise_map).
        exposure_time_map_path : str
            The path to the exposure_time_map .fits file containing the exposure time map \
            (e.g. '/path/to/exposure_time_map.fits')
        exposure_time_map_hdu : int
            The hdu the exposure_time_map is contained in the .fits file specified by *exposure_time_map_path*.
        exposure_time_map_from_single_value : float
            The exposure time of the imaging, which is used to compute the exposure-time map as a single value \
            (see *ExposureTimeMap.from_single_value*).
        exposure_time_map_from_inverse_noise_map : bool
            If True, the exposure-time map is computed from the background noise_map map \
            (see *ExposureTimeMap.from_background_noise_map*)
        background_sky_map_path : str
            The path to the background_sky_map .fits file containing the background sky map \
            (e.g. '/path/to/background_sky_map.fits').
        background_sky_map_hdu : int
            The hdu the background_sky_map is contained in the .fits file specified by *background_sky_map_path*.
        convert_from_electrons : bool
            If True, the input unblurred_image_1d are in unit_label of electrons and all converted to electrons / second using the exposure \
            time map.
        gain : float
            The image gain, used for convert from ADUs.
        convert_from_adus : bool
            If True, the input unblurred_image_1d are in unit_label of adus and all converted to electrons / second using the exposure \
            time map and gain.
        """

        image = abstract_dataset.load_image(
            image_path=image_path, image_hdu=image_hdu, pixel_scales=pixel_scales
        )

        background_noise_map = load_background_noise_map(
            background_noise_map_path=background_noise_map_path,
            background_noise_map_hdu=background_noise_map_hdu,
            pixel_scales=pixel_scales,
            convert_background_noise_map_from_weight_map=convert_background_noise_map_from_weight_map,
            convert_background_noise_map_from_inverse_noise_map=convert_background_noise_map_from_inverse_noise_map,
        )

        if background_noise_map is not None:
            inverse_noise_map = 1.0 / background_noise_map
        else:
            inverse_noise_map = None

        exposure_time_map = abstract_dataset.load_exposure_time_map(
            exposure_time_map_path=exposure_time_map_path,
            exposure_time_map_hdu=exposure_time_map_hdu,
            pixel_scales=pixel_scales,
            shape=image.mask.shape,
            exposure_time=exposure_time_map_from_single_value,
            exposure_time_map_from_inverse_noise_map=exposure_time_map_from_inverse_noise_map,
            inverse_noise_map=inverse_noise_map,
        )

        poisson_noise_map = load_poisson_noise_map(
            poisson_noise_map_path=poisson_noise_map_path,
            poisson_noise_map_hdu=poisson_noise_map_hdu,
            pixel_scales=pixel_scales,
            convert_poisson_noise_map_from_weight_map=convert_poisson_noise_map_from_weight_map,
            convert_poisson_noise_map_from_inverse_noise_map=convert_poisson_noise_map_from_inverse_noise_map,
            image=image,
            exposure_time_map=exposure_time_map,
            poisson_noise_map_from_image=poisson_noise_map_from_image,
            convert_from_electrons=convert_from_electrons,
            gain=gain,
            convert_from_adus=convert_from_adus,
        )

        noise_map = load_noise_map(
            noise_map_path=noise_map_path,
            noise_map_hdu=noise_map_hdu,
            pixel_scales=pixel_scales,
            image=image,
            background_noise_map=background_noise_map,
            exposure_time_map=exposure_time_map,
            convert_noise_map_from_weight_map=convert_noise_map_from_weight_map,
            convert_noise_map_from_inverse_noise_map=convert_noise_map_from_inverse_noise_map,
            noise_map_from_image_and_background_noise_map=noise_map_from_image_and_background_noise_map,
            convert_from_electrons=convert_from_electrons,
            gain=gain,
            convert_from_adus=convert_from_adus,
        )

        if noise_map_non_constant:
            if np.allclose(noise_map, noise_map[0] * np.ones(shape=noise_map.shape)):
                noise_map = noise_map + (
                    0.001 * noise_map[0] * np.random.uniform(size=noise_map.shape_1d)
                )

        if psf_path is not None:

            psf = kernel.Kernel.from_fits(
                file_path=psf_path,
                hdu=psf_hdu,
                pixel_scales=pixel_scales,
                renormalize=renormalize_psf,
            )

        else:

            psf = None

        background_sky_map = load_background_sky_map(
            background_sky_map_path=background_sky_map_path,
            background_sky_map_hdu=background_sky_map_hdu,
            pixel_scales=pixel_scales,
        )

        imaging = Imaging(
            image=image,
            pixel_scales=pixel_scales,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            gain=gain,
            name=lens_name,
        )

        if resized_imaging_shape is not None:
            imaging = imaging.resized_from_new_shape(new_shape=resized_imaging_shape)

        if resized_psf_shape is not None and imaging.psf is not None:
            imaging = imaging.resized_psf_from_new_shape(new_shape=resized_psf_shape)

        if convert_from_electrons:
            imaging = imaging.data_in_electrons()
        elif convert_from_adus:
            imaging = imaging.data_in_adus_from_gain(gain=gain)

        return imaging

    @classmethod
    def simulate(
        cls,
        image,
        exposure_time,
        psf=None,
        exposure_time_map=None,
        background_level=0.0,
        background_sky_map=None,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        name=None,
    ):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Imaging read-out).
        pixel_scales: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        if psf is None:
            psf = kernel.Kernel.no_blur(pixel_scales=image.pixel_scales)
            image_needs_trimming = False
        else:
            image_needs_trimming = True

        if exposure_time_map is None:
            exposure_time_map = arrays.Array.full(
                fill_value=exposure_time, shape_2d=image.shape_2d
            )

        if background_sky_map is None:
            background_sky_map = arrays.Array.full(
                fill_value=background_level, shape_2d=image.shape_2d
            )

        image += background_sky_map

        image = psf.convolved_array_from_array(array=image)

        if image_needs_trimming:
            image = image.trimmed_from_kernel_shape(kernel_shape_2d=psf.shape_2d)
            exposure_time_map = exposure_time_map.trimmed_from_kernel_shape(
                kernel_shape_2d=psf.shape_2d
            )
            background_sky_map = background_sky_map.trimmed_from_kernel_shape(
                kernel_shape_2d=psf.shape_2d
            )

        if add_noise is True:
            noise_realization = generate_poisson_noise(
                image, exposure_time_map, noise_seed
            )
            image += noise_realization
            image_counts = np.multiply(image, exposure_time_map)
            noise_map = np.divide(np.sqrt(image_counts), exposure_time_map)
            noise_map = masked_structures.MaskedArray.manual_1d(
                array=noise_map, mask=noise_map.mask
            )
        else:
            noise_map = arrays.Array.full(
                fill_value=noise_if_add_noise_false,
                shape_2d=image.shape_2d,
                pixel_scales=image.pixel_scales,
            )
            noise_realization = None

        if np.isnan(noise_map).any():
            raise exc.DataException(
                "The noise-map has NaN values in it. This suggests your exposure time and / or"
                "background sky levels are too low, creating signal counts at or close to 0.0."
            )

        image -= background_sky_map

        # ESTIMATE THE BACKGROUND NOISE MAP FROM THE BACKGROUND SKY MAP

        background_noise_map_counts = np.sqrt(
            np.multiply(background_sky_map, exposure_time_map)
        )
        background_noise_map = np.divide(background_noise_map_counts, exposure_time_map)

        # ESTIMATE THE POISSON NOISE MAP FROM THE IMAGE

        image_counts = np.multiply(image, exposure_time_map)
        poisson_noise_map = np.divide(np.sqrt(np.abs(image_counts)), exposure_time_map)

        mask = msk.Mask.unmasked(
            shape_2d=image.shape_2d, pixel_scales=image.pixel_scales
        )

        image = masked_structures.MaskedArray.manual_1d(array=image, mask=mask)
        background_noise_map = masked_structures.MaskedArray.manual_1d(
            array=background_noise_map, mask=mask
        )
        poisson_noise_map = masked_structures.MaskedArray.manual_1d(
            array=poisson_noise_map, mask=mask
        )

        return SimulatedImaging(
            image=image,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            noise_realization=noise_realization,
            name=name,
        )

    def __array_finalize__(self, obj):
        if isinstance(obj, Imaging):
            try:
                self.data = obj.data
                self.psf = obj.psf
                self.noise_map = obj.noise_map
                self.background_noise_map = obj.background_noise_map
                self.poisson_noise_map = obj.poisson_noise_map
                self.exposure_time_map = obj.exposure_time_map
                self.background_sky_map = obj.background_sky_map
                self.origin = obj.origin
            except AttributeError:
                logger.debug(
                    "Original object in Imaging.__array_finalize__ missing one or more attributes"
                )


class SimulatedImaging(Imaging):
    def __init__(
        self,
        image,
        psf,
        noise_map=None,
        background_noise_map=None,
        poisson_noise_map=None,
        exposure_time_map=None,
        background_sky_map=None,
        noise_realization=None,
        name=None,
        **kwargs
    ):

        super(SimulatedImaging, self).__init__(
            image=image,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            name=name,
            kwargs=kwargs,
        )

        self.noise_realization = noise_realization

    def __array_finalize__(self, obj):
        if isinstance(obj, SimulatedImaging):
            try:
                self.data = obj.data
                self.psf = obj.psf
                self.noise_map = obj.noise_map
                self.background_noise_map = obj.background_noise_map
                self.poisson_noise_map = obj.poisson_noise_map
                self.exposure_time_map = obj.exposure_time_map
                self.background_sky_map = obj.background_sky_map
                self.background_noise_realization = obj.background_noise_realization
                self.poisson_noise_realization = obj.poisson_noise_realization
                self.origin = obj.origin
            except AttributeError:
                logger.debug(
                    "Original object in Imaging.__array_finalize__ missing one or more attributes"
                )


def setup_random_seed(seed):
    """Setup the random seed. If the input seed is -1, the code will use a random seed for every run. If it is \
    positive, that seed is used for all runs, thereby giving reproducible results.

    Parameters
    ----------
    seed : int
        The seed of the random number generator.
    """
    if seed == -1:
        seed = np.random.randint(
            0, int(1e9)
        )  # Use one seed, so all regions have identical column non-uniformity.
    np.random.seed(seed)


def generate_poisson_noise(image, exposure_time_map, seed=-1):
    """
    Generate a two-dimensional poisson noise_maps-mappers from an image.

    Values are computed from a Poisson distribution using the image's input values in unit_label of counts.

    Parameters
    ----------
    image : ndarray
        The 2D image, whose values in counts are used to draw Poisson noise_maps values.
    exposure_time_map : Union(ndarray, int)
        2D array of the exposure time in each pixel used to convert to / from counts and electrons per second.
    seed : int
        The seed of the random number generator, used for the random noise_maps maps.

    Returns
    -------
    poisson_noise_map: ndarray
        An array describing simulated poisson noise_maps
    """
    setup_random_seed(seed)
    image_counts = np.multiply(image, exposure_time_map)
    return image - np.divide(
        np.random.poisson(image_counts, image.shape), exposure_time_map
    )


def load_noise_map(
    noise_map_path,
    noise_map_hdu,
    pixel_scales,
    image=None,
    background_noise_map=None,
    exposure_time_map=None,
    convert_noise_map_from_weight_map=False,
    convert_noise_map_from_inverse_noise_map=False,
    noise_map_from_image_and_background_noise_map=False,
    convert_from_electrons=False,
    gain=None,
    convert_from_adus=False,
):
    """Factory for loading the noise-map from a .fits file.

    This factory also includes a number of routines for converting the noise-map from from other unit_label (e.g. \
    a weight map) or computing the noise-map from other unblurred_image_1d (e.g. the imaging image and background noise-map).

    Parameters
    ----------
    noise_map_path : str
        The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')
    noise_map_hdu : int
        The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
    pixel_scales : float
        The size of each pixel in arc seconds.
    image : ndarray
        The image-image, which the noise-map can be calculated using.
    background_noise_map : ndarray
        The background noise-map, which the noise-map can be calculated using.
    exposure_time_map : ndarray
        The exposure-time map, which the noise-map can be calculated using.
    convert_noise_map_from_weight_map : bool
        If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_noise_map_from_inverse_noise_map : bool
        If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
        *NoiseMap.from_inverse_noise_map).
    background_noise_map_path : str
        The path and filename of the .fits image containing the background noise-map.
    background_noise_map_hdu : int
        The hdu the background noise-map is contained in the .fits file that *background_noise_map_path* points too.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    noise_map_from_image_and_background_noise_map : bool
        If True, the noise-map is computed from the observed image and background noise-map \
        (see NoiseMap.from_image_and_background_noise_map).
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in unit_label of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in unit_label of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """
    noise_map_options = sum(
        [
            convert_noise_map_from_weight_map,
            convert_noise_map_from_inverse_noise_map,
            noise_map_from_image_and_background_noise_map,
        ]
    )

    if noise_map_options > 1:
        raise exc.DataException(
            "You have specified more than one method to load the noise_map map, e.g.:"
            "convert_noise_map_from_weight_map | "
            "convert_noise_map_from_inverse_noise_map |"
            "noise_map_from_image_and_background_noise_map"
        )

    if noise_map_options == 0 and noise_map_path is not None:
        return arrays.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu, pixel_scales=pixel_scales
        )
    elif convert_noise_map_from_weight_map and noise_map_path is not None:
        weight_map = arrays.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu, pixel_scales=pixel_scales
        )
        return data_converter.noise_map_from_weight_map(weight_map=weight_map)
    elif convert_noise_map_from_inverse_noise_map and noise_map_path is not None:
        inverse_noise_map = arrays.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu, pixel_scales=pixel_scales
        )
        return data_converter.noise_map_from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map
        )
    elif noise_map_from_image_and_background_noise_map:

        if background_noise_map is None:
            raise exc.DataException(
                "Cannot compute the noise-map from the image and background noise_map map if a "
                "background noise_map map is not supplied."
            )

        if (
            not (convert_from_electrons or convert_from_adus)
            and exposure_time_map is None
        ):
            raise exc.DataException(
                "Cannot compute the noise-map from the image and background noise_map map if an "
                "exposure-time (or exposure time map) is not supplied to convert to adus"
            )

        if convert_from_adus and gain is None:
            raise exc.DataException(
                "Cannot compute the noise-map from the image and background noise_map map if a"
                "gain is not supplied to convert from adus"
            )

        return data_converter.noise_map_from_image_and_background_noise_map(
            image=image,
            background_noise_map=background_noise_map,
            exposure_time_map=exposure_time_map,
            convert_from_electrons=convert_from_electrons,
            gain=gain,
            convert_from_adus=convert_from_adus,
        )
    else:
        raise exc.DataException(
            "A noise_map map was not loaded, specify a noise_map_path or option to compute a noise_map map."
        )


def load_background_noise_map(
    background_noise_map_path,
    background_noise_map_hdu,
    pixel_scales,
    convert_background_noise_map_from_weight_map=False,
    convert_background_noise_map_from_inverse_noise_map=False,
):
    """Factory for loading the background noise-map from a .fits file.

    This factory also includes a number of routines for converting the background noise-map from from other unit_label (e.g. \
    a weight map).

    Parameters
    ----------
    background_noise_map_path : str
        The path to the background_noise_map .fits file containing the background noise-map \
        (e.g. '/path/to/background_noise_map.fits')
    background_noise_map_hdu : int
        The hdu the background_noise_map is contained in the .fits file specified by *background_noise_map_path*.
    pixel_scales : float
        The size of each pixel in arc seconds.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    """
    background_noise_map_options = sum(
        [
            convert_background_noise_map_from_weight_map,
            convert_background_noise_map_from_inverse_noise_map,
        ]
    )

    if background_noise_map_options == 0 and background_noise_map_path is not None:
        return arrays.Array.from_fits(
            file_path=background_noise_map_path,
            hdu=background_noise_map_hdu,
            pixel_scales=pixel_scales,
        )
    elif (
        convert_background_noise_map_from_weight_map
        and background_noise_map_path is not None
    ):
        weight_map = arrays.Array.from_fits(
            file_path=background_noise_map_path,
            hdu=background_noise_map_hdu,
            pixel_scales=pixel_scales,
        )
        return data_converter.noise_map_from_weight_map(weight_map=weight_map)
    elif (
        convert_background_noise_map_from_inverse_noise_map
        and background_noise_map_path is not None
    ):
        inverse_noise_map = arrays.Array.from_fits(
            file_path=background_noise_map_path,
            hdu=background_noise_map_hdu,
            pixel_scales=pixel_scales,
        )
        return data_converter.noise_map_from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map
        )
    else:
        return None


def load_poisson_noise_map(
    poisson_noise_map_path,
    poisson_noise_map_hdu,
    pixel_scales,
    convert_poisson_noise_map_from_weight_map=False,
    convert_poisson_noise_map_from_inverse_noise_map=False,
    poisson_noise_map_from_image=False,
    image=None,
    exposure_time_map=None,
    convert_from_electrons=False,
    gain=None,
    convert_from_adus=False,
):
    """Factory for loading the Poisson noise-map from a .fits file.

    This factory also includes a number of routines for converting the Poisson noise-map from from other unit_label (e.g. \
    a weight map) or computing the Poisson noise_map from other unblurred_image_1d (e.g. the imaging image).

    Parameters
    ----------
    poisson_noise_map_path : str
        The path to the poisson_noise_map .fits file containing the Poisson noise-map \
         (e.g. '/path/to/poisson_noise_map.fits')
    poisson_noise_map_hdu : int
        The hdu the poisson_noise_map is contained in the .fits file specified by *poisson_noise_map_path*.
    pixel_scales : float
        The size of each pixel in arc seconds.
    convert_poisson_noise_map_from_weight_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_poisson_noise_map_from_inverse_noise_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    poisson_noise_map_from_image : bool
        If True, the Poisson noise-map is estimated using the image.
    image : ndarray
        The image, which the Poisson noise-map can be calculated using.
    background_noise_map : ndarray
        The background noise-map, which the Poisson noise-map can be calculated using.
    exposure_time_map : ndarray
        The exposure-time map, which the Poisson noise-map can be calculated using.
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in unit_label of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in unit_label of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """
    poisson_noise_map_options = sum(
        [
            convert_poisson_noise_map_from_weight_map,
            convert_poisson_noise_map_from_inverse_noise_map,
            poisson_noise_map_from_image,
        ]
    )

    if poisson_noise_map_options == 0 and poisson_noise_map_path is not None:
        return arrays.Array.from_fits(
            file_path=poisson_noise_map_path,
            hdu=poisson_noise_map_hdu,
            pixel_scales=pixel_scales,
        )
    elif poisson_noise_map_from_image:

        if (
            not (convert_from_electrons or convert_from_adus)
            and exposure_time_map is None
        ):
            raise exc.DataException(
                "Cannot compute the Poisson noise-map from the image if an "
                "exposure-time (or exposure time map) is not supplied to convert to adus"
            )

        if convert_from_adus and gain is None:
            raise exc.DataException(
                "Cannot compute the Poisson noise-map from the image if a"
                "gain is not supplied to convert from adus"
            )

        return data_converter.poisson_noise_map_from_image_and_exposure_time_map(
            image=image,
            exposure_time_map=exposure_time_map,
            convert_from_electrons=convert_from_electrons,
            gain=gain,
            convert_from_adus=convert_from_adus,
        )

    elif (
        convert_poisson_noise_map_from_weight_map and poisson_noise_map_path is not None
    ):
        weight_map = arrays.Array.from_fits(
            file_path=poisson_noise_map_path,
            hdu=poisson_noise_map_hdu,
            pixel_scales=pixel_scales,
        )
        return data_converter.noise_map_from_weight_map(weight_map=weight_map)
    elif (
        convert_poisson_noise_map_from_inverse_noise_map
        and poisson_noise_map_path is not None
    ):
        inverse_noise_map = arrays.Array.from_fits(
            file_path=poisson_noise_map_path,
            hdu=poisson_noise_map_hdu,
            pixel_scales=pixel_scales,
        )
        return data_converter.noise_map_from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map
        )
    else:
        return None


def load_background_sky_map(
    background_sky_map_path, background_sky_map_hdu, pixel_scales
):
    """Factory for loading the background sky from a .fits file.

    Parameters
    ----------
    background_sky_map_path : str
        The path to the background_sky_map .fits file containing the background sky map \
        (e.g. '/path/to/background_sky_map.fits').
    background_sky_map_hdu : int
        The hdu the background_sky_map is contained in the .fits file specified by *background_sky_map_path*.
    pixel_scales : float
        The size of each pixel in arc seconds.
    """
    if background_sky_map_path is not None:
        return arrays.Array.from_fits(
            file_path=background_sky_map_path,
            hdu=background_sky_map_hdu,
            pixel_scales=pixel_scales,
        )
    else:
        return None



from autoarray.plot import plotters
from autoarray.plot import inversion_plots, mat_objs


@plotters.set_include_and_sub_plotter
@plotters.set_subplot_filename
def subplot_fit_interferometer(fit, include=None, sub_plotter=None):

    number_subplots = 6

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    residual_map_vs_uv_distances(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    normalized_residual_map_vs_uv_distances(
        fit=fit, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=3)

    chi_squared_map_vs_uv_distances(fit=fit, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=4)

    residual_map_vs_uv_distances(
        fit=fit, plot_real=False, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=5)

    normalized_residual_map_vs_uv_distances(
        fit=fit, plot_real=False, include=include, plotter=sub_plotter
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=6)

    chi_squared_map_vs_uv_distances(
        fit=fit, plot_real=False, include=include, plotter=sub_plotter
    )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


def individuals(
    fit,
    plot_visibilities=False,
    plot_noise_map=False,
    plot_signal_to_noise_map=False,
    plot_model_visibilities=False,
    plot_residual_map=False,
    plot_normalized_residual_map=False,
    plot_chi_squared_map=False,
    include=None,
    plotter=None,
):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    if plot_visibilities:

        visibilities(fit=fit, include=include, plotter=plotter)

    if plot_noise_map:

        noise_map(fit=fit, include=include, plotter=plotter)

    if plot_signal_to_noise_map:

        signal_to_noise_map(fit=fit, include=include, plotter=plotter)

    if plot_model_visibilities:

        model_visibilities(fit=fit, include=include, plotter=plotter)

    if plot_residual_map:

        residual_map_vs_uv_distances(
            fit=fit, plot_real=True, include=include, plotter=plotter
        )

        residual_map_vs_uv_distances(
            fit=fit, plot_real=False, include=include, plotter=plotter
        )

    if plot_normalized_residual_map:

        normalized_residual_map_vs_uv_distances(
            fit=fit, plot_real=True, include=include, plotter=plotter
        )

        normalized_residual_map_vs_uv_distances(
            fit=fit, plot_real=False, include=include, plotter=plotter
        )

    if plot_chi_squared_map:

        chi_squared_map_vs_uv_distances(
            fit=fit, plot_real=True, include=include, plotter=plotter
        )

        chi_squared_map_vs_uv_distances(
            fit=fit, plot_real=False, include=include, plotter=plotter
        )


@plotters.set_include_and_plotter
@plotters.set_labels
def visibilities(fit, include=None, plotter=None):
    """Plot the visibilities of a lens fit.

    Set *autolens.datas.grid.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    visibilities : datas.imaging.datas.Imaging
        The datas-datas, which include the observed datas, noise_map-map, PSF, signal-to-noise_map-map, etc.
    origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """
    plotter.plot_grid(grid=fit.visibilities)


@plotters.set_include_and_plotter
@plotters.set_labels
def noise_map(fit, include=None, plotter=None):
    """Plot the noise-map of a lens fit.

    Set *autolens.datas.grid.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    visibilities : datas.imaging.datas.Imaging
        The datas-datas, which include the observed datas, noise_map-map, PSF, signal-to-noise_map-map, etc.
    origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """
    plotter.plot_grid(grid=fit.visibilities, color_array=fit.noise_map[:, 0])


@plotters.set_include_and_plotter
@plotters.set_labels
def signal_to_noise_map(fit, include=None, plotter=None):
    """Plot the noise-map of a lens fit.

    Set *autolens.datas.grid.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    visibilities : datas.imaging.datas.Imaging
    The datas-datas, which include the observed datas, signal_to_noise_map-map, PSF, signal-to-signal_to_noise_map-map, etc.
    origin : True
    If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """
    plotter.plot_grid(grid=fit.visibilities, color_array=fit.signal_to_noise_map[:, 0])


@plotters.set_include_and_plotter
@plotters.set_labels
def model_visibilities(fit, include=None, plotter=None):
    """Plot the model visibilities of a fit.

    Set *autolens.datas.grid.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which include a list of every model visibilities, residual_map, chi-squareds, etc.
    visibilities_index : int
        The index of the datas in the datas-set of which the model visibilities is plotted.
    """
    plotter.plot_grid(grid=fit.visibilities)


@plotters.set_include_and_plotter
@plotters.set_labels
def residual_map_vs_uv_distances(
    fit,
    plot_real=True,
    label_yunits="V$_{R,data}$ - V$_{R,model}$",
    label_xunits=r"UV$_{distance}$ (k$\lambda$)",
    include=None,
    plotter=None,
):
    """Plot the residual-map of a lens fit.

    Set *autolens.datas.grid.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which include a list of every model visibilities, residual_map, chi-squareds, etc.
    visibilities_index : int
        The index of the datas in the datas-set of which the residual_map are plotted.
    """

    if plot_real:
        y = fit.residual_map[:, 0]
        plotter = plotter.plotter_with_new_labels(title=plotter.labels.title + " Real")
        plotter = plotter.plotter_with_new_output(
            filename=plotter.output.filename + "_real"
        )
    else:
        y = fit.residual_map[:, 1]
        plotter = plotter.plotter_with_new_labels(title=plotter.labels.title + " Imag")
        plotter = plotter.plotter_with_new_output(
            filename=plotter.output.filename + "_imag"
        )

    plotter.plot_line(
        y=y,
        x=fit.masked_interferometer.interferometer.uv_distances / 10 ** 3.0,
        plot_axis_type="scatter",
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def normalized_residual_map_vs_uv_distances(
    fit,
    plot_real=True,
    label_yunits="V$_{R,data}$ - V$_{R,model}$",
    label_xunits=r"UV$_{distance}$ (k$\lambda$)",
    include=None,
    plotter=None,
):
    """Plot the residual-map of a lens fit.

    Set *autolens.datas.grid.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which include a list of every model visibilities, residual_map, chi-squareds, etc.
    visibilities_index : int
        The index of the datas in the datas-set of which the residual_map are plotted.
    """

    if plot_real:
        y = fit.residual_map[:, 0]
        plotter = plotter.plotter_with_new_labels(title=plotter.labels.title + " Real")
        plotter = plotter.plotter_with_new_output(
            filename=plotter.output.filename + "_real"
        )
    else:
        y = fit.residual_map[:, 1]
        plotter = plotter.plotter_with_new_labels(title=plotter.labels.title + " Imag")
        plotter = plotter.plotter_with_new_output(
            filename=plotter.output.filename + "_imag"
        )

    plotter.plot_line(
        y=y,
        x=fit.masked_interferometer.interferometer.uv_distances / 10 ** 3.0,
        plot_axis_type="scatter",
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def chi_squared_map_vs_uv_distances(
    fit,
    plot_real=True,
    label_yunits="V$_{R,data}$ - V$_{R,model}$",
    label_xunits=r"UV$_{distance}$ (k$\lambda$)",
    include=None,
    plotter=None,
):
    """Plot the residual-map of a lens fit.

    Set *autolens.datas.grid.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which include a list of every model visibilities, residual_map, chi-squareds, etc.
    visibilities_index : int
        The index of the datas in the datas-set of which the residual_map are plotted.
    """

    if plot_real:
        y = fit.residual_map[:, 0]
        plotter = plotter.plotter_with_new_labels(title=plotter.labels.title + " Real")
        plotter = plotter.plotter_with_new_output(
            filename=plotter.output.filename + "_real"
        )
    else:
        y = fit.residual_map[:, 1]
        plotter = plotter.plotter_with_new_labels(title=plotter.labels.title + " Imag")
        plotter = plotter.plotter_with_new_output(
            filename=plotter.output.filename + "_imag"
        )

    plotter.plot_line(
        y=y,
        x=fit.masked_interferometer.interferometer.uv_distances / 10 ** 3.0,
        plot_axis_type="scatter",
    )



import autoarray as aa
import autoarray.plot as aplt
import pytest
import os

from os import path

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_plot_path_setup():
    return "{}/../../test_files/plotting/fit_interferometer/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


@pytest.fixture(autouse=True)
def set_config_path():
    aa.conf.instance = aa.conf.Config(
        path.join(directory, "../test_files/plot"), path.join(directory, "output")
    )


def test__fit_quantities_are_output(fit_interferometer_7, plot_path, plot_patch):

    aplt.fit_interferometer.visibilities(
        fit=fit_interferometer_7,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert plot_path + "visibilities.png" in plot_patch.paths

    aplt.fit_interferometer.noise_map(
        fit=fit_interferometer_7,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert plot_path + "noise_map.png" in plot_patch.paths

    aplt.fit_interferometer.signal_to_noise_map(
        fit=fit_interferometer_7,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert plot_path + "signal_to_noise_map.png" in plot_patch.paths

    aplt.fit_interferometer.model_visibilities(
        fit=fit_interferometer_7,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert plot_path + "model_visibilities.png" in plot_patch.paths

    aplt.fit_interferometer.residual_map_vs_uv_distances(
        fit=fit_interferometer_7,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert plot_path + "residual_map_vs_uv_distances_real.png" in plot_patch.paths

    aplt.fit_interferometer.residual_map_vs_uv_distances(
        fit=fit_interferometer_7,
        plot_real=False,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert plot_path + "residual_map_vs_uv_distances_imag.png" in plot_patch.paths

    aplt.fit_interferometer.normalized_residual_map_vs_uv_distances(
        fit=fit_interferometer_7,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert (
        plot_path + "normalized_residual_map_vs_uv_distances_real.png"
        in plot_patch.paths
    )

    aplt.fit_interferometer.normalized_residual_map_vs_uv_distances(
        fit=fit_interferometer_7,
        plot_real=False,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert (
        plot_path + "normalized_residual_map_vs_uv_distances_imag.png"
        in plot_patch.paths
    )

    aplt.fit_interferometer.chi_squared_map_vs_uv_distances(
        fit=fit_interferometer_7,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert plot_path + "chi_squared_map_vs_uv_distances_real.png" in plot_patch.paths

    aplt.fit_interferometer.chi_squared_map_vs_uv_distances(
        fit=fit_interferometer_7,
        plot_real=False,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="png")),
    )

    assert plot_path + "chi_squared_map_vs_uv_distances_imag.png" in plot_patch.paths


def test__fit_sub_plot(fit_interferometer_7, plot_path, plot_patch):

    aplt.fit_interferometer.subplot_fit_interferometer(
        fit=fit_interferometer_7,
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "subplot_fit_interferometer.png" in plot_patch.paths


def test__fit_individuals__dependent_on_input(
    fit_interferometer_7, plot_path, plot_patch
):

    aplt.fit_interferometer.individuals(
        fit=fit_interferometer_7,
        plot_visibilities=True,
        plot_noise_map=False,
        plot_signal_to_noise_map=False,
        plot_model_visibilities=True,
        plot_chi_squared_map=True,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "visibilities.png" in plot_patch.paths

    assert plot_path + "noise_map.png" not in plot_patch.paths

    assert plot_path + "signal_to_noise_map.png" not in plot_patch.paths

    assert plot_path + "model_visibilities.png" in plot_patch.paths

    assert plot_path + "residual_map_vs_uv_distances_real.png" not in plot_patch.paths

    assert plot_path + "residual_map_vs_uv_distances_imag.png" not in plot_patch.paths

    assert (
        plot_path + "normalized_residual_map_vs_uv_distances_real.png"
        not in plot_patch.paths
    )

    assert (
        plot_path + "normalized_residual_map_vs_uv_distances_imag.png"
        not in plot_patch.paths
    )

    assert plot_path + "chi_squared_map_vs_uv_distances_real.png" in plot_patch.paths

    assert plot_path + "chi_squared_map_vs_uv_distances_imag.png" in plot_patch.paths
